@@ -4,7 +4,6 @@ This module does not import anything from the library except during type checkin
 """
 import datetime
 import uuid
-import warnings
 from copy import deepcopy
 from dataclasses import dataclass, field, fields
 from pathlib import Path
@@ -16,7 +15,6 @@ from typing_extensions import Protocol, runtime
 from .helpers import dump, get_full_type_name, load, unique_name
 from .java import to_dict, to_java
 
-
 # from .actors.thread_actors import ThreadProxy
 
 if TYPE_CHECKING:
@@ -101,7 +99,7 @@ class MechanismState:
         return MechanismState(**self.__dict__)
 
     def __deepcopy__(self, memodict={}):
-        d = {k: deepcopy(v) for k, v in self.__dict__.items()}
+        d = {k: deepcopy(v, memo=memodict) for k, v in self.__dict__.items()}
         return MechanismState(**d)
 
     def __getitem__(self, item):
@@ -140,17 +138,25 @@ class AgentMechanismInterface:
     """The time limit in seconds for this negotiation session. None indicates infinity"""
     step_time_limit: float
     """The time limit in seconds for each step of this negotiation session. None indicates infinity"""
+    negotiator_time_limit: float
+    """The time limit in seconds to wait for negotiator responses of this negotiation session. None indicates infinity"""
     n_steps: int
     """The allowed number of steps for this negotiation. None indicates infinity"""
     dynamic_entry: bool
     """Whether it is allowed for agents to enter/leave the negotiation after it starts"""
     max_n_agents: int
     """Maximum allowed number of agents in the session. None indicates no limit"""
+    imap: Dict[Union[str, int], Union[str, int]]
+    """A map that translates issue names to indices and issue indices to names"""
     annotation: Dict[str, Any] = field(default_factory=dict)
     """An arbitrary annotation as a `Dict[str, Any]` that is always available for all agents"""
-
     _mechanism = None
 
+    @property
+    def params(self):
+        """Returns the parameters used to initialize the mechanism."""
+        return self._mechanism.params
+
     def random_outcomes(
         self, n: int = 1, astype: Type["Outcome"] = None
     ) -> List["Outcome"]:
@@ -231,6 +237,26 @@ class AgentMechanismInterface:
         """Syntactic sugar for state.n_negotiators"""
         return self.state.n_negotiators
 
+    @property
+    def negotiator_ids(self) -> List[str]:
+        """Gets the IDs of all negotiators"""
+        return self._mechanism.negotiator_ids
+
+    @property
+    def negotiator_names(self) -> List[str]:
+        """Gets the namess of all negotiators"""
+        return self._mechanism.negotiator_names
+
+    @property
+    def agent_ids(self) -> List[str]:
+        """Gets the IDs of all agents owning all negotiators"""
+        return self._mechanism.agent_ids
+
+    @property
+    def agent_names(self) -> List[str]:
+        """Gets the names of all agents owning all negotiators"""
+        return self._mechanism.agent_names
+
     def __str__(self):
         return str(self.__dict__)
 
@@ -252,7 +278,9 @@ class AgentMechanismInterface:
         return AgentMechanismInterface(**self.__dict__)
 
     def __deepcopy__(self, memodict={}):
-        d = {k: deepcopy(v) for k, v in self.__dict__.items()}
+        d = {k: deepcopy(v, memo=memodict) for k, v in self.__dict__.items()}
+        if "_mechanism" in d.keys():
+            del d["_mechanism"]
         return AgentMechanismInterface(**d)
 
     def __getitem__(self, item):
@@ -271,8 +299,7 @@ class AgentMechanismInterface:
 
 
 class _ShadowAgentMechanismInterface:
-    """Used to represent an AMI to Java.
-    """
+    """Used to represent an AMI to Java."""
 
     def randomOutcomes(self, n: int):
         return to_java(self.shadow.random_outcomes(n))
@@ -323,7 +350,7 @@ class Runnable(Protocol):
         pass
 
 
-class NamedObject(object):
+class NamedObject:
     """The base class of all named entities.
 
     All named entities need to call this class's __init__() somewhere during initialization.
@@ -331,13 +358,20 @@ class NamedObject(object):
     Args:
         name (str): The given name of the entity. Notice that the class will add this to a  base that depends
                     on the child's class name.
+        id (str): A unique identifier in the whole system. In principle you should let the system create
+                  this identifier by passing None. In special cases like in serialization you may want to
+                  set the id directly
 
     """
 
-    def __init__(self, name: str = None) -> None:
+    def __init__(self, name: str = None, *, id: str = None) -> None:
         if name is not None:
             name = str(name)
-        self.__uuid = (f"{name}-" if name is not None else "") + str(uuid.uuid4())
+        self.__uuid = (
+            (f"{name}-" if name is not None else "") + str(uuid.uuid4())
+            if not id
+            else id
+        )
         if name is None or len(name) == 0:
             name = unique_name("", add_time=False, rand_digits=16)
         self.__name = name
@@ -570,8 +604,10 @@ class Rational(NamedObject):
         ufun: An optional utility function to attach to the object
     """
 
-    def __init__(self, name: str = None, ufun: Optional["UtilityFunction"] = None):
-        super().__init__(name)
+    def __init__(
+        self, name: str = None, ufun: Optional["UtilityFunction"] = None, id: str = None
+    ):
+        super().__init__(name, id=id)
         self._utility_function = ufun
         self._init_utility = ufun
         self._ufun_modified = ufun is not None
