@@ -1,26 +1,22 @@
 import copy
 import time
 from abc import abstractmethod
-
-import numpy as np
 from typing import Callable, List, Optional, Tuple, Union
 
-from .expectors import MeanExpector
-from .common import _locs, _uppers
+import numpy as np
+
 from ..common import AgentMechanismInterface, MechanismState
 from ..modeling import AdaptiveDiscreteAcceptanceModel
 from ..outcomes import Outcome, ResponseType
-from ..sao import (
-    AspirationNegotiator,
-    SAONegotiator,
-)
+from ..sao import AspirationNegotiator, SAONegotiator
 from ..utilities import (
     IPUtilityFunction,
     MappingUtilityFunction,
     UtilityDistribution,
-    UtilityFunction,
     UtilityValue,
 )
+from .common import _locs, _uppers
+from .expectors import MeanExpector
 
 __all__ = ["BaseElicitor"]
 
@@ -502,7 +498,7 @@ class BaseElicitor(SAONegotiator):
         if len(self.offerable_outcomes) == 0:
             return None, self.reserved_value
         best, best_utility, bsf = None, self.reserved_value, self.reserved_value
-        for i, outcome in enumerate(self.offerable_outcomes):
+        for outcome in self.offerable_outcomes:
             if outcome is None:
                 continue
             utilitiy = self.offering_utility(outcome, state=state)
@@ -557,7 +553,6 @@ class BaseElicitor(SAONegotiator):
 
     def before_eliciting(self) -> None:
         """Called by apply just before continuously calling elicit_single"""
-        pass
 
     @abstractmethod
     def can_elicit(self) -> bool:
