@@ -1,11 +1,35 @@
 import logging
 import math
 import random
-from pprint import pformat
-
-import pandas as pd
+import warnings
 from typing import Any, Dict
 
+import pandas as pd
+
+from ..genius import GeniusNegotiator
+from ..helpers import create_loggers, instantiate
+from ..inout import load_genius_domain_from_folder
+from ..mechanisms import Mechanism
+from ..modeling import UncertainOpponentModel
+from ..negotiators import AspirationMixin
+from ..outcomes import Outcome
+from ..sao import (
+    AspirationNegotiator,
+    LimitedOutcomesAcceptor,
+    LimitedOutcomesNegotiator,
+    OnlyBestNegotiator,
+    RandomNegotiator,
+    SAOMechanism,
+    SAOState,
+    ToughNegotiator,
+)
+from ..utilities import (
+    IPUtilityFunction,
+    MappingUtilityFunction,
+    UtilityDistribution,
+    UtilityFunction,
+    UtilityValue,
+)
 from .baseline import DummyElicitor, FullKnowledgeElicitor
 from .expectors import BalancedExpector, MaxExpector, MinExpector
 from .pandora import FullElicitor, RandomElicitor
@@ -25,30 +49,6 @@ from .voi import (
     np,
     time,
 )
-from ..genius import GeniusNegotiator
-from ..helpers import create_loggers, instantiate
-from ..inout import load_genius_domain_from_folder
-from ..mechanisms import Mechanism
-from ..modeling import UncertainOpponentModel
-from ..negotiators import AspirationMixin
-from ..outcomes import Outcome
-from ..sao import (
-    AspirationNegotiator,
-    LimitedOutcomesAcceptor,
-    OnlyBestNegotiator,
-    RandomNegotiator,
-    SAOState,
-    ToughNegotiator,
-    SAOMechanism,
-    LimitedOutcomesNegotiator,
-)
-from ..utilities import (
-    IPUtilityFunction,
-    MappingUtilityFunction,
-    UtilityDistribution,
-    UtilityFunction,
-    UtilityValue,
-)
 
 __all__ = ["SAOElicitingMechanism"]
 
@@ -82,7 +82,9 @@ def create_negotiator(
                 **kwargs,
             )
     elif negotiator_type == "random":
-        negotiator = RandomNegotiator(can_propose=can_propose,)
+        negotiator = RandomNegotiator(
+            can_propose=can_propose,
+        )
     elif negotiator_type == "tough":
         negotiator = ToughNegotiator(can_propose=can_propose)
     elif negotiator_type in ("only_best", "best_only", "best"):
@@ -111,7 +113,9 @@ def create_negotiator(
                 toughness = 1 - 0.9 * toughness
             asp_kind = toughness
         negotiator = AspirationNegotiator(
-            aspiration_type=asp_kind, can_propose=can_propose, **kwargs,
+            aspiration_type=asp_kind,
+            can_propose=can_propose,
+            **kwargs,
         )
     elif negotiator_type.startswith("genius"):
         class_name = negotiator_type[len("genius") :]
@@ -189,7 +193,7 @@ class SAOElicitingMechanism(SAOMechanism):
                 "Must specify a folder to run from or a number of outcomes"
             )
         if genius_folder is not None:
-            domain, agent_info, issues = load_genius_domain_from_folder(
+            domain, agent_info, _ = load_genius_domain_from_folder(
                 folder_name=genius_folder,
                 force_single_issue=True,
                 keep_issue_names=False,
@@ -399,7 +403,7 @@ class SAOElicitingMechanism(SAOMechanism):
                 elif "optimal" in type_:
                     prune = "prune" in type_ or "fast" in type_
                     if "no" in type_:
-                        no_prune = not prune
+                        prune = not prune
                     return VOIOptimalElicitor(
                         user=user,
                         resolution=resolution,
@@ -562,13 +566,22 @@ class SAOElicitingMechanism(SAOMechanism):
         self.elicitation_state["n_queries"] = 0
         return True
 
-    def plot(self, consider_costs=False):
+    def plot(
+        self,
+        visible_negotiators=(0, 1),
+        plot_utils=True,
+        plot_outcomes=False,
+        utility_range=None,
+        consider_costs=False,
+    ):
         try:
-            import matplotlib.pyplot as plt
             import matplotlib.gridspec as gridspec
+            import matplotlib.pyplot as plt
 
             if len(self.negotiators) > 2:
-                print("Cannot visualize negotiations with more than 2 negotiators")
+                warnings.warn(
+                    "Cannot visualize negotiations with more than 2 negotiators"
+                )
             else:
                 # has_front = int(len(self.outcomes[0]) <2)
                 has_front = 1
@@ -578,7 +591,8 @@ class SAOElicitingMechanism(SAOMechanism):
                 history["relative_time"] = [_[0].relative_time for _ in self.history]
                 history["step"] = [_[0].step for _ in self.history]
                 history = history.loc[~history.offer.isnull(), :]
-                ufuns = self._get_ufuns(consider_costs=consider_costs)
+                # ufuns = self._get_ufuns(consider_costs=consider_costs)
+                ufuns = self._get_ufuns()
                 elicitor_dist = self.negotiators[1].utility_function
                 outcomes = self.outcomes
 
@@ -828,4 +842,3 @@ class SAOElicitingMechanism(SAOMechanism):
             self.elicitation_state["total_voi"] = self.negotiators[1].total_voi
         else:
             self.elicitation_state["total_voi"] = None
-
