@@ -3,32 +3,25 @@ import random
 import time
 from heapq import heapify, heappop
 from math import sqrt
+from typing import Callable, List, Optional, Tuple, Union
 
 import numpy as np
 import scipy.optimize as opt
-from typing import Callable, List, Optional, Tuple, Union
 
-from .common import _loc, _scale
-from .base import BaseElicitor
-from .expectors import (
-    MeanExpector,
-    MinExpector,
-    MaxExpector,
-    AspiringExpector,
-    BalancedExpector,
-)
-from ..common import AgentMechanismInterface, MechanismState
+from ..common import MechanismState
 from ..modeling import AdaptiveDiscreteAcceptanceModel
 from ..negotiators import AspirationMixin
 from ..outcomes import Outcome
-from ..sao import (
-    AspirationNegotiator,
-    SAONegotiator,
-)
-from ..utilities import (
-    IPUtilityFunction,
-    UtilityDistribution,
-    UtilityValue,
+from ..sao import AspirationNegotiator, SAONegotiator
+from ..utilities import UtilityValue
+from .base import BaseElicitor
+from .common import _loc, _scale
+from .expectors import (
+    AspiringExpector,
+    BalancedExpector,
+    MaxExpector,
+    MeanExpector,
+    MinExpector,
 )
 
 __all__ = [
@@ -162,7 +155,7 @@ class BasePandoraElicitor(BaseElicitor, AspirationMixin):
         return self.aspiration(state.relative_time)
 
     def update_cutoff_utility(self) -> None:
-        """
+        r"""
         Updates the cutoff utility under which no elicitation is done.
 
         Remarks:
@@ -292,6 +285,8 @@ class BasePandoraElicitor(BaseElicitor, AspirationMixin):
               the outocme with maximum estimated utility and the negative
               of the corresponding estimated utility.
         """
+        if self.unknown is None:
+            self.init_unknowns()
         unknowns = self.unknown
         if len(unknowns) > 0:
             return -unknowns[0][0], unknowns[0][1]
@@ -302,6 +297,8 @@ class BasePandoraElicitor(BaseElicitor, AspirationMixin):
         Updates the unknown list (and makes sure it is a heap) given the given utility
         value for the given outcome.
         """
+        if self.unknown is None:
+            self.init_unknowns()
         self.unknown[0] = (
             -weitzman_index_uniform(_loc(u), _scale(u), self.user.cost_of_asking()),
             self.unknown[0][1],
@@ -318,13 +315,17 @@ class BasePandoraElicitor(BaseElicitor, AspirationMixin):
         """
         Removes the best offer from the unkonwn list and returns it.
         """
+        if self.unknown is None:
+            self.init_unknowns()
         return heappop(self.unknown)
 
     def can_elicit(self) -> bool:
         """
         Checks whether there are any unknowns in the unknowns list.
         """
-        return len(self.unknown) != 0
+        if self.unknown is None:
+            self.init_unknowns()
+        return self.unknown and len(self.unknown) != 0
 
     def z_index(self, updated_outcomes: Optional[List[Outcome]] = None):
         """
@@ -582,7 +583,12 @@ class PandoraElicitor(BasePandoraElicitor):
        (https://dl.acm.org/doi/10.5555/2832249.2832250).
     """
 
-    def __init__(self, strategy: "EStrategy", user: "User", **kwargs,) -> None:
+    def __init__(
+        self,
+        strategy: "EStrategy",
+        user: "User",
+        **kwargs,
+    ) -> None:
         kwargs.update(
             dict(
                 base_negotiator=AspirationNegotiator(),
@@ -602,7 +608,9 @@ class PandoraElicitor(BasePandoraElicitor):
             )
         )
         super().__init__(
-            strategy=strategy, user=user, **kwargs,
+            strategy=strategy,
+            user=user,
+            **kwargs,
         )
 
 
@@ -617,7 +625,6 @@ class FastElicitor(PandoraElicitor):
 
     def update_best_offer_utility(self, outcome: "Outcome", u: UtilityValue):
         """We need not do anything here as we will remove the outcome anyway to the known list"""
-        pass
 
     def do_elicit(self, outcome: "Outcome", state: MechanismState):
         return self.expect(super().do_elicit(outcome, state), state=state)
@@ -630,10 +637,17 @@ class OptimalIncrementalElicitor(FastElicitor):
     by changes in the opponent model.
     """
 
-    def __init__(self, strategy: "EStrategy", user: "User", **kwargs,) -> None:
+    def __init__(
+        self,
+        strategy: "EStrategy",
+        user: "User",
+        **kwargs,
+    ) -> None:
         kwargs.update(dict(incremental=True))
         super().__init__(
-            strategy=strategy, user=user, **kwargs,
+            strategy=strategy,
+            user=user,
+            **kwargs,
         )
 
 
@@ -641,10 +655,17 @@ class MeanElicitor(OptimalIncrementalElicitor):
     """Same as `OptimalIncrementalElicitor` using `MeanExpector` for
     estimating utilities"""
 
-    def __init__(self, strategy: "EStrategy", user: "User", **kwargs,) -> None:
+    def __init__(
+        self,
+        strategy: "EStrategy",
+        user: "User",
+        **kwargs,
+    ) -> None:
         kwargs.update(dict(expector_factory=MeanExpector))
         super().__init__(
-            strategy=strategy, user=user, **kwargs,
+            strategy=strategy,
+            user=user,
+            **kwargs,
         )
 
 
@@ -652,10 +673,17 @@ class BalancedElicitor(OptimalIncrementalElicitor):
     """Same as `OptimalIncrementalElicitor` using `MeanExpector` for
     estimating utilities"""
 
-    def __init__(self, strategy: "EStrategy", user: "User", **kwargs,) -> None:
+    def __init__(
+        self,
+        strategy: "EStrategy",
+        user: "User",
+        **kwargs,
+    ) -> None:
         kwargs.update(dict(expector_factory=BalancedExpector))
         super().__init__(
-            strategy=strategy, user=user, **kwargs,
+            strategy=strategy,
+            user=user,
+            **kwargs,
         )
 
 
@@ -696,10 +724,17 @@ class PessimisticElicitor(OptimalIncrementalElicitor):
     """Same as `OptimalIncrementalElicitor` using the minimum to estimate
     utilities."""
 
-    def __init__(self, strategy: "EStrategy", user: "User", **kwargs,) -> None:
+    def __init__(
+        self,
+        strategy: "EStrategy",
+        user: "User",
+        **kwargs,
+    ) -> None:
         kwargs.update(dict(expector_factory=MinExpector))
         super().__init__(
-            strategy=strategy, user=user, **kwargs,
+            strategy=strategy,
+            user=user,
+            **kwargs,
         )
 
 
@@ -707,8 +742,15 @@ class OptimisticElicitor(OptimalIncrementalElicitor):
     """Same as `OptimalIncrementalElicitor` using the maximum to estimate
     utilities."""
 
-    def __init__(self, strategy: "EStrategy", user: "User", **kwargs,) -> None:
+    def __init__(
+        self,
+        strategy: "EStrategy",
+        user: "User",
+        **kwargs,
+    ) -> None:
         kwargs.update(dict(expector_factory=MaxExpector))
         super().__init__(
-            strategy=strategy, user=user, **kwargs,
+            strategy=strategy,
+            user=user,
+            **kwargs,
         )
