@@ -8,6 +8,7 @@ from abc import abstractmethod
 from collections import defaultdict
 from heapq import heapify, heappop, heappush
 from warnings import warn
+
 import numpy as np
 
 try:
@@ -26,19 +27,16 @@ except ImportError:
     )
 from typing import Callable, List, Optional, Tuple, Union
 
-from .base import BaseElicitor
-from .common import _scale, argmax
-from .expectors import Expector, MeanExpector
-from .queries import Query, Answer, RangeConstraint
-from .strategy import EStrategy
 from ..common import MechanismState
 from ..modeling import AdaptiveDiscreteAcceptanceModel
 from ..outcomes import Outcome
-from ..sao import (
-    AspirationNegotiator,
-    SAONegotiator,
-)
+from ..sao import AspirationNegotiator, SAONegotiator
 from ..utilities import UtilityValue
+from .base import BaseElicitor
+from .common import _scale, argmax
+from .expectors import Expector, MeanExpector
+from .queries import Answer, Query, RangeConstraint
+from .strategy import EStrategy
 
 __all__ = [
     "BaseVOIElicitor",
@@ -101,7 +99,9 @@ class BaseVOIElicitor(BaseElicitor):
         **kwargs,
     ) -> None:
         super().__init__(
-            strategy=strategy, user=user, **kwargs,
+            strategy=strategy,
+            user=user,
+            **kwargs,
         )
         self.eeu_query = None
         self.query_index_of_outcome = None
@@ -120,6 +120,7 @@ class BaseVOIElicitor(BaseElicitor):
         self,
         ufun: Optional[Union["IPUtilityFunction", "UtilityDistribution"]],
         queries: Optional[List[Query]] = None,
+        **kwargs,
     ) -> None:
         """
         Initializes the elicitation process once.
@@ -168,7 +169,6 @@ class BaseVOIElicitor(BaseElicitor):
                 for i, (_o, _q, _c) in enumerate(self.queries):
                     queries_of_outcome[_o].append(i)
                 self.queries_of_outcome = queries_of_outcome
-            pass
         self.init_query_eeus()
         self._elicitation_time += time.perf_counter() - strt_time
 
@@ -213,7 +213,6 @@ class BaseVOIElicitor(BaseElicitor):
 
     def before_eliciting(self):
         """Called every round before trying to elicit. Does nothing"""
-        pass
 
     def on_opponent_model_updated(
         self, outcomes: List[Outcome], old: List[float], new: List[float]
@@ -382,13 +381,13 @@ class BaseVOIElicitor(BaseElicitor):
     def add_query(self, qeeu: Tuple[float, int]) -> None:
         """Adds a query to the heap of queries
 
-            Args:
-                qeeu: A Tuple giving (-EEU, query_index)
+        Args:
+            qeeu: A Tuple giving (-EEU, query_index)
 
-            Remarks:
-                - Note that the first member of the tuple is **minus** the EEU
-                - The sedond member of the tuple is an index of the query in
-                  the queries list (not the query itself).
+        Remarks:
+            - Note that the first member of the tuple is **minus** the EEU
+            - The sedond member of the tuple is an index of the query in
+              the queries list (not the query itself).
         """
         heappush(self.eeu_query, qeeu)
 
@@ -438,7 +437,7 @@ class VOIElicitor(BaseVOIElicitor):
 
     def eeu(self, policy: np.ndarray, eus: np.ndarray) -> float:
         """Expected Expected Negotiator for following the policy"""
-        p = np.ones((len(policy) + 1))
+        p = np.ones(len(policy) + 1)
         m = self.opponent_model.acceptance_probabilities()[policy]
         r = 1 - m
         eup = -eus * m
@@ -491,7 +490,7 @@ class VOIElicitor(BaseVOIElicitor):
             if best_index is not None:
                 indices.remove(best_index)
         self.outcome_in_policy = {}
-        for i, (eu, outcome) in enumerate(eu_policy):
+        for i, (_, outcome) in enumerate(eu_policy):
             self.outcome_in_policy[outcome] = i
         heapify(eu_policy)
         self.eu_policy, self.current_eeu = eu_policy, best_eeu
@@ -516,7 +515,7 @@ class VOIElicitor(BaseVOIElicitor):
             _policy = np.array([_[1] for _ in eu_policy])
             _eus = np.array([_[0] for _ in eu_policy])
             answer_eeus.append(self.eeu(policy=_policy, eus=_eus))
-        return cost - sum([a * b for a, b in zip(answer_probabilities, answer_eeus)])
+        return cost - sum(a * b for a, b in zip(answer_probabilities, answer_eeus))
 
 
 class VOIFastElicitor(BaseVOIElicitor):
@@ -539,7 +538,7 @@ class VOIFastElicitor(BaseVOIElicitor):
         eu_policy = sortedlist(zip(eus, range(n_outcomes)))
         policy = np.array([_[1] for _ in eu_policy])
         eu = np.array([_[0] for _ in eu_policy])
-        p = np.ones((len(policy) + 1))
+        p = np.ones(len(policy) + 1)
         ac = self.opponent_model.acceptance_probabilities()[policy]
         eup = -eu * ac
         r = 1 - ac
@@ -626,7 +625,7 @@ class VOIFastElicitor(BaseVOIElicitor):
                 self.eu_policy.remove(new_util)
             answer_eeus.append(reeu)
         self.eu_policy.add(old_util)
-        qeeu = cost - sum([a * b for a, b in zip(answer_probabilities, answer_eeus)])
+        qeeu = cost - sum(a * b for a, b in zip(answer_probabilities, answer_eeus))
         return qeeu
 
 
@@ -636,7 +635,7 @@ class VOINoUncertaintyElicitor(BaseVOIElicitor):
 
     def eeu(self, policy: np.ndarray, eup: np.ndarray) -> float:
         """Expected Expected Negotiator for following the policy"""
-        p = np.ones((len(policy) + 1))
+        p = np.ones(len(policy) + 1)
         r = 1 - self.opponent_model.acceptance_probabilities()[policy]
         p[1:] = np.cumprod(r)
         try:
@@ -667,7 +666,8 @@ class VOINoUncertaintyElicitor(BaseVOIElicitor):
         )
         self.eu_policy = eu_policy
         self.outcome_in_policy = {}
-        for j, (_, indx) in enumerate(eu_policy):
+        # for j, (_, indx) in enumerate(eu_policy):
+        for _, indx in eu_policy:
             self.outcome_in_policy[indx] = (_, indx)
 
     def init_query_eeus(self) -> None:
@@ -823,7 +823,7 @@ class VOIOptimalElicitor(BaseElicitor):
             < self.reserved_value
         ):
             return False
-        outcome, query, cost = self.queries[q]
+        outcome, query, _ = self.queries[q]
         if query is None:
             return False
         self.queries[q] = (None, None, None)
@@ -910,7 +910,7 @@ class VOIOptimalElicitor(BaseElicitor):
                 ):  # ignore cases where it is impossible to go to this large j
                     continue
                 try:
-                    sjm1, sjm = s[jm - 1] if jm > 0 else 0.0, s[jm]
+                    _, sjm = s[jm - 1] if jm > 0 else 0.0, s[jm]
                     if m1 > 1e-6:
                         y = ((sk1 - sk) + m * (sjm - sk1)) / m1
                     else:
@@ -1012,7 +1012,7 @@ class VOIOptimalElicitor(BaseElicitor):
         eu_policy = sortedlist(zip(eus, range(n_outcomes)))
         policy = np.array([_[1] for _ in eu_policy])
         eu = np.array([_[0] for _ in eu_policy])
-        p = np.ones((len(policy) + 1))
+        p = np.ones(len(policy) + 1)
         ac = self.opponent_model.acceptance_probabilities()[policy]
         eup = -eu * ac
         r = 1 - ac
