@@ -1,16 +1,22 @@
 """Implements Event management"""
-import itertools
+import json
 import random
+import time
+import warnings
 from collections import defaultdict
 from dataclasses import dataclass
-from typing import Any, Callable, Dict, List, Optional, Set, Union
+from pathlib import Path
+from typing import Any, Callable, Dict, Iterable, List, Optional, Union
 
 from .common import NamedObject
+from .outcomes import Issue
+from .serialization import serialize
 
 __all__ = [
     "Event",
     "EventSource",
     "EventSink",
+    "EventLogger",
     "Notification",
     "Notifier",
     "Notifiable",
@@ -19,6 +25,8 @@ __all__ = [
 
 @dataclass
 class Event:
+    """An event that can be raised and consumed"""
+
     __slots__ = ["type", "data"]
     type: str
     data: Any
@@ -29,26 +37,98 @@ class EventSource:
 
     def __init__(self):
         super().__init__()
-        self.__sinks: Dict[str, list] = defaultdict(list)
+        self.__sinks: Dict[Optional[str], List[EventSink]] = defaultdict(list)
 
     def announce(self, event: Event):
-        """Raises an event and informs all event sinks that are registerd for notifications
+        """Raises an event and informs all event sinks that are registered for notifications
         on this event type"""
 
-        sinks = self.__sinks.get(event.type, [])
+        sinks = self.__sinks.get(event.type, []) + self.__sinks.get(None, [])
         random.shuffle(sinks)
         for sink in sinks:
             sink.on_event(event=event, sender=self)
 
-    def register_listener(self, event_type: str, listener: "EventSink"):
+    def register_listener(self, event_type: Optional[str], listener: "EventSink"):
+        """Registers a listener for the given event_type.
+
+        Args:
+            event_type: The type to register. If None, the listener will
+                        be registered for all types
+            listener: The listening agent (must have an `on_event` method
+                      that receives an event: `Event` and a sender: `EventSource`)
+        """
         self.__sinks[event_type].append(listener)
 
 
 class EventSink:
+    """An object capable of receiving events"""
+
     def on_event(self, event: Event, sender: EventSource):
         pass
 
 
+def myvars(x):
+    if not x:
+        return x
+    return {k: v for k in dir(x) if not k.startswith("_") and not k.endswith("_")}
+
+
+class EventLogger(EventSink):
+    """
+    Logs events to a file
+
+    Args:
+        file_name: Name of the file to save events into
+        types: The types of events to save. If None, all events will be saved
+    """
+
+    def __init__(self, file_name: Union[str, Path], types: List[str] = None):
+        file_name = Path(file_name)
+        file_name.parent.mkdir(parents=True, exist_ok=True)
+        self._file_name = file_name
+        self._types = set(types) if types else None
+        self._start = time.perf_counter()
+
+    def reset_timer(self):
+        self._start = time.perf_counter()
+
+    def on_event(self, event: Event, sender: EventSource):
+        if not self._file_name:
+            return
+        if self._types is not None and event.type not in self._types:
+            return
+
+        def _simplify(x):
+            if x is None:
+                return None
+            if isinstance(x, (str, int, float)):
+                return x
+            if isinstance(x, Issue):
+                return dict(name=x.name, values=x.values)
+            if isinstance(x, dict):
+                return {k: _simplify(v) for k, v in x.items()}
+            for y in ("id", "name"):
+                if hasattr(x, y):
+                    return getattr(x, y)
+            if isinstance(x, Iterable):
+                return list(_simplify(_) for _ in x)
+            return str(x)
+            # return _simplify(myvars(x))
+
+        try:
+            sid = sender.id if hasattr(sender, "id") else serialize(sender)
+            d = dict(
+                sender=sid,
+                time=time.perf_counter() - self._start,
+                type=event.type,
+                data=_simplify(event.data),
+            )
+            with open(self._file_name, "a") as f:
+                f.write(f"{json.dumps(d)},\n")
+        except Exception as e:
+            warnings.warn(f"Failed to log {str(event)}: {str(e)}")
+
+
 @dataclass
 class Notification:
     __slots__ = ["type", "data"]
@@ -57,11 +137,15 @@ class Notification:
 
 
 class Notifier(NamedObject):
+    """An object that can notify other objects"""
+
     def notify(self, notifiable: "Notifiable", notification: Notification):
         notifiable.on_notification_(notification=notification, notifier=self.id)
 
 
 class Notifiable:
+    """An object that can be notified"""
+
     def add_handler(
         self, notification_type: str, callback: Callable[[Notification, str], bool]
     ):
@@ -92,7 +176,7 @@ class Notifiable:
         """
         try:
             return self.__notification_handlers[notification_type]
-        except (ValueError, IndexError, AttributeError) as e:
+        except (ValueError, IndexError, AttributeError):
             return []
 
     def remove_handler(
@@ -133,7 +217,7 @@ class Notifiable:
                 if callback(notification, notifier):
                     break
             return True
-        except (IndexError, ValueError, AttributeError) as e:
+        except (IndexError, ValueError, AttributeError):
             self.on_notification(notification, notifier)
             return False
 
@@ -149,4 +233,3 @@ class Notifiable:
 
             - override this method to provide a catch-all notification handling method.
         """
-        pass
