@@ -4,22 +4,32 @@
 This set of utlities can be extended but must be backward compatible for at
 least two versions
 """
+import atexit
+import base64
+import concurrent
 import copy
 import datetime
 import importlib
+import itertools
 import json
 import logging
 import math
 import os
 import pathlib
-import dill as pickle
 import random
 import re
+import socket
 import string
 import sys
-from collections import defaultdict
-from enum import Enum
 import traceback
+import warnings
+from collections import defaultdict
+from concurrent.futures import TimeoutError
+from concurrent.futures.thread import ThreadPoolExecutor
+from contextlib import contextmanager
+from enum import Enum
+from pathlib import Path
+from types import FunctionType, LambdaType
 from typing import (
     TYPE_CHECKING,
     Any,
@@ -34,12 +44,14 @@ from typing import (
 )
 
 import colorlog
+import dill as pickle
 import inflect
 import numpy as np
 import pandas as pd
 import scipy.stats as stats
 import stringcase
 import yaml
+from scipy.stats import tmean
 
 from negmas.config import NEGMAS_CONFIG
 from negmas.generics import GenericMapping, IterableMapping, gmap, ikeys
@@ -48,6 +60,12 @@ if TYPE_CHECKING:
     pass
 
 __all__ = [
+    "get_one_float",
+    "get_one_int",
+    "make_range",
+    "TimeoutError",
+    "TimeoutCaller",
+    "PATH",
     "shortest_unique_names",
     "create_loggers",
     # 'MultiIssueUtilityFunctionMapping',
@@ -72,9 +90,19 @@ __all__ = [
     "add_records",
     "load",
     "exception2str",
+    "is_jsonable",
+    "is_lambda_function",
+    "is_type",
+    "is_non_lambda_function",
+    "force_single_thread",
+    "single_thread",
+    "TYPE_START",
+    "truncated_mean",
 ]
 # conveniently named classes
-
+TYPE_START = "__TYPE__:"
+BYTES_START = "__BYTES__:"
+PATH_START = "__PATH__:"
 """Maps from a single issue to a Negotiator function."""
 # MultiIssueUtilityFunctionMapping = Union[
 #    Callable[['Issues'], 'UtilityFunction'], Mapping['Issues', 'UtilityFunction']]  # type: ignore
@@ -96,6 +124,26 @@ LOGS_BASE_DIR = "./logs"
 
 DEFAULT_DUMP_EXTENSION = NEGMAS_CONFIG.get("default_dump_extension", "json")
 
+PATH = Union[pathlib.Path, str]
+
+
+def get_one_int(i: Union[int, Tuple[int, int]]):
+    if isinstance(i, int):
+        return i
+    return random.randint(*i)
+
+
+def get_one_float(rng: Union[float, Tuple[float, float]]):
+    if isinstance(rng, float):
+        return rng
+    return random.random() * (rng[1] - rng[0]) + rng[0]
+
+
+def make_range(x: Union[Any, Tuple[Any, Any]]) -> Tuple[Any, Any]:
+    if isinstance(x, Iterable):
+        return x
+    return (x, x)
+
 
 class ReturnCause(Enum):
     TIMEOUT = 0
@@ -103,38 +151,79 @@ class ReturnCause(Enum):
     FAILURE = 2
 
 
-def shortest_unique_names(strs: List[str], sep="."):
+def shortest_unique_names(
+    strs: List[str], sep=".", max_compression=False, guarantee_unique=False
+):
     """
     Finds the shortest unique strings starting from the end of each input
     string based on the separator.
 
     The final strings will only be unique if the inputs are unique.
 
+    Args:
+        strs: A list of strings
+        sep: The separator used to separate fields in each string
+        max_compression: If True, each string will be further compressed
+                         by taking the shortest prefix that keeps the
+                         strings unique (if they were originally unique)
+        guarantee_unique: If given, random characters will be postfixed on
+                         strings to guarantee uniquness
+
     Example:
-        given ["a.b.c", "d.e.f", "a.d.c"] it will generate ["b.c", "f", "d.c"]
+        given ["a.b.cat", "d.e.f", "a.d.cat"] it will generate ["b.c", "f", "d.cat"]
+        if max_compression was false and will generate ["b", "f", "d"] if it was
+        True
     """
+    if len(strs) < 2:
+        return strs
+    if guarantee_unique and len(set(strs)) != len(strs):
+        chars = string.digits + string.ascii_letters
+        for i in range(len(strs) - 1):
+            others = set(strs[:i] + strs[i + 1 :])
+            while strs[i] in others:
+                for a in chars:
+                    if strs[i] + a not in others:
+                        strs[i] = strs[i] + a
+                        break
+                else:
+                    strs[i] = strs[i] + unique_name("", False, 1, "")
+
     lsts = [_.split(sep) for _ in strs]
     names = [_[-1] for _ in lsts]
-    if len(names) == len(set(names)):
-        return names
-    locs = defaultdict(list)
-    for i, s in enumerate(names):
-        locs[s].append(i)
-    mapping = {"": ""}
-    for s, l in locs.items():
-        if len(s) < 1:
-            continue
-        if len(l) == 1:
-            mapping[strs[l[0]]] = s
-            continue
-        strs_new = [sep.join(lsts[_][:-1]) for _ in l]
-        prefixes = shortest_unique_names(strs_new, sep)
-        for loc, prefix in zip(l, prefixes):
-            x = sep.join([prefix, s])
-            if x.startswith(sep):
-                x = x[len(sep) :]
-            mapping[strs[loc]] = x
-    return [mapping[_] for _ in strs]
+    if len(names) != len(set(names)):
+        locs = defaultdict(list)
+        for i, s in enumerate(names):
+            locs[s].append(i)
+        mapping = {"": ""}
+        for s, l in locs.items():
+            if len(s) < 1:
+                continue
+            if len(l) == 1:
+                mapping[strs[l[0]]] = s
+                continue
+            strs_new = [sep.join(lsts[_][:-1]) for _ in l]
+            prefixes = shortest_unique_names(
+                strs_new, sep, max_compression, guarantee_unique
+            )
+            for loc, prefix in zip(l, prefixes):
+                x = sep.join([prefix, s])
+                if x.startswith(sep):
+                    x = x[len(sep) :]
+                mapping[strs[loc]] = x
+        strs = [mapping[_] for _ in strs]
+    else:
+        strs = names
+    if not max_compression:
+        return strs
+    for i, s in enumerate(strs):
+        for j in range(1, len(s)):
+            for k in itertools.chain(range(i), range(i + 1, len(strs))):
+                if strs[k][:j] == s[:j]:
+                    break
+            else:
+                strs[i] = s[:j]
+                break
+    return strs
 
 
 def create_loggers(
@@ -179,15 +268,18 @@ def create_loggers(
         if len(logger.handlers) > 0:
             return logger
         logger.setLevel(logging.DEBUG)
+    else:
+        logger = logging.getLogger()
     # create formatter
-    if colored and "colorlog" in sys.modules and os.isatty(2):
+    file_formatter = logging.Formatter(format_str)
+    if colored and "colorlog" in sys.modules and os.isatty(2) and screen_level:
         date_format = "%Y-%m-%d %H:%M:%S"
         cformat = "%(log_color)s" + format_str
-        formatter = colorlog.ColoredFormatter(
+        screen_formatter = colorlog.ColoredFormatter(
             cformat,
             date_format,
             log_colors={
-                "DEBUG": "reset",
+                "DEBUG": "magenta",
                 "INFO": "green",
                 "WARNING": "yellow",
                 "ERROR": "red",
@@ -195,13 +287,13 @@ def create_loggers(
             },
         )
     else:
-        formatter = logging.Formatter(format_str)
+        screen_formatter = logging.Formatter(format_str)
     if screen_level is not None and (module_wide_log_file or app_wide_log_file):
         # create console handler and set level to logdebug
         screen_logger = logging.StreamHandler()
         screen_logger.setLevel(screen_level)
         # add formatter to ch
-        screen_logger.setFormatter(formatter)
+        screen_logger.setFormatter(screen_formatter)
         # add ch to logger
         logger.addHandler(screen_logger)
     if file_name is not None and file_level is not None:
@@ -226,13 +318,13 @@ def create_loggers(
         os.makedirs(os.path.dirname(file_name), exist_ok=True)  # type: ignore
         file_logger = logging.FileHandler(file_name)
         file_logger.setLevel(file_level)
-        file_logger.setFormatter(formatter)
+        file_logger.setFormatter(file_formatter)
         logger.addHandler(file_logger)
     return logger
 
 
 def snake_case(s: str) -> str:
-    """ Converts a string from CamelCase to snake_case
+    """Converts a string from CamelCase to snake_case
 
     Example:
 
@@ -255,7 +347,7 @@ def snake_case(s: str) -> str:
 def camel_case(
     s: str, capitalize_first: bool = False, lower_first: bool = False
 ) -> str:
-    """ Converts a string from snake_case to CamelCase
+    """Converts a string from snake_case to CamelCase
 
     Example:
 
@@ -290,7 +382,11 @@ def camel_case(
 
 
 def unique_name(
-    base: Union[pathlib.Path, str], add_time=True, rand_digits=8, sep="/"
+    base: Union[pathlib.Path, str],
+    add_time=True,
+    add_host=False,
+    rand_digits=8,
+    sep="/",
 ) -> str:
     """Return a unique name.
 
@@ -313,13 +409,14 @@ def unique_name(
 
     """
     _time, rand_part = "", ""
+    host_part = socket.gethostname() if add_host else ""
     if rand_digits > 0:
         rand_part = "".join(
             random.choices(string.digits + string.ascii_letters, k=rand_digits)
         )
     if add_time:
         _time = datetime.datetime.now().strftime("%Y%m%dH%H%M%S%f")
-    sub = _time + rand_part
+    sub = _time + host_part + rand_part
     if len(sub) == 0:
         return base
     if len(base) == 0:
@@ -340,18 +437,18 @@ def is_nonzero_file(fpath: str) -> bool:
 def pretty_string(src: Any, tab_size=2, compact=False) -> str:
     """Recursively print nested elements.
 
-        Args:
-            src (Any): The source to be converted to a printable string
-            tab_size (int): Tab size in spaces
-            compact (bool): If true the output is  converted into a single line
+    Args:
+        src (Any): The source to be converted to a printable string
+        tab_size (int): Tab size in spaces
+        compact (bool): If true the output is  converted into a single line
 
-        Returns:
-            str: The pretty version of the input
+    Returns:
+        str: The pretty version of the input
 
-        Remarks:
-            - This function assumes that the patterns `` "`` and ``":`` do not appear anywhere in the input.
-              If they appear, the space, : will be removed.
-        """
+    Remarks:
+        - This function assumes that the patterns `` "`` and ``":`` do not appear anywhere in the input.
+          If they appear, the space, : will be removed.
+    """
     s = _pretty_string(src, dpth=0, current_key="", tab_size=tab_size)
     if compact:
         return s.replace("\n", "")
@@ -388,13 +485,13 @@ def _pretty_string(src, dpth=0, current_key="", tab_size=2) -> str:
         output += tabs(dpth) + "]"
     else:
         if len(current_key) > 0:
-            output += tabs(dpth) + '"%s":%s' % (current_key, src)
+            output += tabs(dpth) + f'"{current_key}":{src}'
         else:
             output += tabs(dpth) + "%s" % src
     return output
 
 
-class LazyInitializable(object):
+class LazyInitializable:
     """Base Negotiator for all agents
 
     Supports a set_params function that can be used for lazy initialization
@@ -474,10 +571,9 @@ class LazyInitializable(object):
               this object.
             - You should **never** call `adjust_params()` directly anywhere.
         """
-        pass
 
 
-class Distribution(object):
+class Distribution:
     """Any distribution from scipy.stats with overloading of addition and multiplication.
 
     Args:
@@ -569,18 +665,17 @@ class Distribution(object):
     def __or__(self, other):
         if isinstance(other, float) or isinstance(other, int):
             return float(other)
-        if self.dtype == "uniform":
-            raise NotImplementedError(
-                "Current implementation assumes an overlap otherwise a mixture must be returned"
-            )
-            beg = min(self.loc, other.loc)
-            end = max(self.scale + self.loc, other.loc + other.scale)
-            return Distribution(self.dtype, loc=beg, scale=end - beg)
+        # if self.dtype == "uniform":
+        #     raise NotImplementedError(
+        #         "Current implementation assumes an overlap otherwise a mixture must be returned"
+        #     )
+        # beg = min(self.loc, other.loc)
+        # end = max(self.scale + self.loc, other.loc + other.scale)
+        # return Distribution(self.dtype, loc=beg, scale=end - beg)
         raise NotImplementedError()
 
     def prob(self, val: float) -> float:
-        """Returns the probability for the given value
-        """
+        """Returns the probability for the given value"""
         return self.dist.prob(val)
 
     def sample(self, size: int = 1) -> np.ndarray:
@@ -770,7 +865,7 @@ class ConfigReader:
                             config = str(name)
                         else:
                             raise ValueError(f"Cannot find config in {config}.")
-            with open(config, "r") as f:
+            with open(config) as f:
                 if config.endswith(".json"):
                     config = json.load(f)
                 elif config.endswith(".cfg"):
@@ -957,12 +1052,21 @@ def import_by_name(full_name: str) -> Any:
 
 
 def get_class(
-    class_name: Union[str, Type], module_name: str = None, scope: dict = None
+    class_name: Union[str, Type],
+    module_name: str = None,
+    scope: dict = None,
+    allow_nonstandard_names=False,
 ) -> Type:
     """Imports and creates a class object for the given class name"""
     if not isinstance(class_name, str):
         return class_name
+
+    # remove explicit type annotation in the string. Used when serializing
+    while class_name.startswith(TYPE_START):
+        class_name = class_name[len(TYPE_START) :]
+
     modules: List[str] = []
+
     if module_name is not None:
         modules = module_name.split(".")
     modules += class_name.split(".")
@@ -970,7 +1074,10 @@ def get_class(
         raise ValueError(
             f"Cannot get the class {class_name} in module {module_name}  (modules {modules})"
         )
-    class_name = stringcase.pascalcase(modules[-1])
+    if not class_name.startswith("builtins") or allow_nonstandard_names:
+        class_name = stringcase.pascalcase(modules[-1])
+    else:
+        class_name = modules[-1]
     if len(modules) < 2:
         return eval(class_name, scope)
     module_name = ".".join(modules[:-1])
@@ -1011,6 +1118,14 @@ def humanize_time(secs, align=False, always_show_all_units=False):
     return ":".join(parts)
 
 
+def is_jsonable(x):
+    try:
+        json.dumps(x)
+        return True
+    except:
+        return False
+
+
 class NpEncoder(json.JSONEncoder):
     def default(self, obj):
         if isinstance(obj, np.integer):
@@ -1019,12 +1134,47 @@ class NpEncoder(json.JSONEncoder):
             return float(obj)
         elif isinstance(obj, np.ndarray):
             return obj.tolist()
+        elif isinstance(obj, bytes):
+            encoded = base64.b64encode(
+                obj
+            )  # b'ZGF0YSB0byBiZSBlbmNvZGVk' (notice the "b")
+            return BYTES_START + encoded.decode("ascii")  #
+        elif isinstance(obj, Path):
+            return PATH_START + str(obj)
+        elif not is_jsonable(obj):
+            # it may be a type. Always convert types to full names when saving to json
+            try:
+                obj = TYPE_START + get_full_type_name(obj)
+            except:
+                return obj
+            return obj
         else:
-            return super(NpEncoder, self).default(obj)
+            return super().default(obj)
+
+
+class NpDecorder(json.JSONDecoder):
+    def default(self, obj):
+        if isinstance(obj, str):
+            if obj.startswith(BYTES_START):
+                return base64.b64decode(
+                    obj[BYTES_START:]
+                )  # b'ZGF0YSB0byBiZSBlbmNvZGVk' (notice the "b")
+            elif obj.startswith(TYPE_START):
+                return get_class(
+                    obj[TYPE_START:]
+                )  # b'ZGF0YSB0byBiZSBlbmNvZGVk' (notice the "b")
+            elif obj.startswith(PATH_START):
+                return Path(
+                    obj[PATH_START:]
+                )  # b'ZGF0YSB0byBiZSBlbmNvZGVk' (notice the "b")
+        return super().default(obj)
 
 
 def dump(
-    d: Any, file_name: Union[str, os.PathLike, pathlib.Path], sort_keys=True
+    d: Any,
+    file_name: Union[str, os.PathLike, pathlib.Path],
+    sort_keys=True,
+    compact=False,
 ) -> None:
     """
     Saves an object depending on the extension of the file given. If the filename given has no extension,
@@ -1033,7 +1183,8 @@ def dump(
     Args:
         d: Object to save
         file_name: file name
-        sort_keys: If true, the keys will be sorted before saving for JSON
+        sort_keys: If true, the keys will be sorted before saving
+        compact: If given, a compact representation will be tried
 
     Remarks:
 
@@ -1051,7 +1202,13 @@ def dump(
             pass
     if file_name.suffix == ".json":
         with open(file_name, "w") as f:
-            json.dump(d, f, sort_keys=sort_keys, indent=2, cls=NpEncoder)
+            json.dump(
+                d,
+                f,
+                sort_keys=sort_keys,
+                indent=2 if not compact else None,
+                cls=NpEncoder,
+            )
     elif file_name.suffix == ".yaml":
         with open(file_name, "w") as f:
             yaml.safe_dump(d, f)
@@ -1091,10 +1248,10 @@ def load(file_name: Union[str, os.PathLike, pathlib.Path]) -> Any:
         return d
 
     if file_name.suffix == ".json":
-        with open(file_name, "r") as f:
+        with open(file_name) as f:
             d = json.load(f)
     elif file_name.suffix == ".yaml":
-        with open(file_name, "r") as f:
+        with open(file_name) as f:
             yaml.safe_load(f)
     elif file_name.suffix == ".pickle":
         with open(file_name, "rb") as f:
@@ -1106,8 +1263,31 @@ def load(file_name: Union[str, os.PathLike, pathlib.Path]) -> Any:
     return d
 
 
+def is_lambda_function(obj):
+    """Checks if the given object is a lambda function"""
+    return isinstance(obj, LambdaType) and obj.__name__ == "<lambda>"
+
+
+def is_type(obj):
+    """Checks if the given object is a type converted to string"""
+    return isinstance(obj, Type)
+
+
+def is_not_type(obj):
+    """Checks if the given object is not a type converted to string"""
+    return not is_type(obj)
+
+
+def is_non_lambda_function(obj):
+    """Checks if the given object is not a lambda function"""
+    return isinstance(obj, FunctionType) and obj.__name__ != "<lambda>"
+
+
 def add_records(
-    file_name: Union[str, os.PathLike], data: Any, col_names: Optional[List[str]] = None
+    file_name: Union[str, os.PathLike],
+    data: Any,
+    col_names: Optional[List[str]] = None,
+    raise_exceptions=False,
 ) -> None:
     """
     Adds records to a csv file
@@ -1117,6 +1297,7 @@ def add_records(
         file_name: file name
         data: data to use for creating the record
         col_names: Names in the data.
+        raise_exceptions: If given, exceptions  are raised on failure
 
     Returns:
 
@@ -1141,23 +1322,26 @@ def add_records(
     file_name.parent.mkdir(parents=True, exist_ok=True)
     new_file = True
     mode = "a"
-    if file_name.exists():
+    # if file_name.exists():
+    if is_nonzero_file(file_name):
         new_file = False
-        with open(file_name, "r") as f:
+        with open(file_name) as f:
             header = f.readline().strip().strip("\n")
         cols = header.split(",")
         for col in cols:
             if len(col) > 0 and col not in data.columns:
                 data[col] = None
-        if set([_ for _ in data.columns]) == set(cols):
+        if {_ for _ in data.columns} == set(cols):
             data = data.loc[:, cols]
         else:
             try:
                 old_data = pd.read_csv(file_name, index_col=None)
                 data = pd.concat((old_data, data), axis=0, ignore_index=True)
-            except Exception:
-                print(
-                    f"Failed to read data from file {str(file_name)} will override it"
+            except Exception as e:
+                if raise_exceptions:
+                    raise e
+                warnings.warn(
+                    f"Failed to read data from file {str(file_name)} will override it\n{e}"
                 )
 
             mode = "w"
@@ -1168,3 +1352,173 @@ def add_records(
 
 def exception2str(limit=None, chain=True) -> str:
     return traceback.format_exc(limit=limit, chain=chain)
+
+
+def truncated_mean(
+    scores: np.ndarray,
+    limits: Optional[Tuple[float, float]] = None,
+    top_limit=2.0,
+    bottom_limit=float("inf"),
+    base="tukey",
+    return_limits=False,
+) -> Union[float, Tuple[float, Optional[Tuple[float, float]]]]:
+    """
+    Calculates the truncated mean
+
+    Args:
+        scores: A list of scores for which to calculate the truncated mean
+        limits: The limits to use for trimming the scores. If not given, they will
+                be calculated based on `top_limit`, `bottom_limit` and `base.`
+                You can pass the special value "mean" as a string to disable limits and
+                calcualte the mean. You can pass the special value "median" to calculate
+                the median (which is the same as passing top_limit==bottom_limit=0.5
+                and base == "scores").
+        top_limit: top limit on scores to use for truncated mean calculation. See `base`
+        bottom_limit: bottom limit on scores to use for truncated mean calculation. See `base`
+        base: The base for calculating the limits used to apply the `top_limit` and `bottom_limit`.
+              Possible values are:
+              - zscore: the number of sigmas to remove above/below. A good default choice is 3. Pass inf to disable a side.
+              - tukey: the fraction of IQR to remove above/below. A good default choice is 1.5 or 3 (we use 2). Pass inf to disable a side.
+              - iqr : same as tukey
+              - iqr_fraction: the fraction is interpreted as the fraction of scores above/below the 1st/3rd qauntile
+              - scores: the fraction is interpreted as fraction of highest and lowest scores
+              - fraction: the fraction is interpreted as literal fraction of the values (i.e. given 10 values and 0.1, removes 1 value)
+              - mean: simply returns the mean (limits ignored)
+              - median: simply returns the median (limits ignored)
+        return_limits: If true, the method will also return the limiting scores used in its mean calculation.
+    """
+
+    scores = np.asarray(scores)
+    scores = scores[~np.isnan(scores)]
+
+    if isinstance(limits, str) and limits.lower() == "mean":
+        return tmean(scores, None) if not return_limits else (tmean(scores, None), None)
+    if isinstance(limits, str) and limits.lower() == "median":
+        return np.median(scores) if not return_limits else (np.median(scores), None)
+    if limits is not None:
+        return np.mean(scores) if not return_limits else (np.mean(scores), None)
+
+    if base == "zscore":
+        m, s = np.nanmean(scores), np.nanstd(scores)
+        limits = (m - s * bottom_limit, m + s * top_limit)
+    elif base in ("tukey", "iqr"):
+        q1, q3 = np.quantile(scores, 0.25), np.quantile(scores, 0.75)
+        iqr = q3 - q1
+        limits = (
+            q1 - (bottom_limit * iqr if not np.isinf(bottom_limit) else bottom_limit),
+            q3 + (top_limit * iqr if not np.isinf(top_limit) else top_limit),
+        )
+    elif base == "iqr_fraction":
+        bottom_limit = min(1, max(0, bottom_limit))
+        top_limit = min(1, max(0, top_limit))
+        limits = (np.quantile(scores, 0.25), np.quantile(scores, 0.75))
+        high = np.sort(scores[scores > limits[1]])
+        low = np.sort(scores[scores < limits[0]])
+        limits = (
+            low[int((len(low) - 1) * bottom_limit)] if len(low) > 0 else None,
+            high[int((len(high) - 1) * (1 - top_limit))] if len(high) > 0 else None,
+        )
+    elif base == "fraction":
+        bottom_limit = min(1, max(0, bottom_limit))
+        top_limit = min(1, max(0, top_limit))
+        scores = np.sort(scores)
+        top_indx = int((len(scores) - 1) * (1 - top_limit))
+        bottom_indx = int((len(scores) - 1) * bottom_limit)
+        if top_indx < bottom_indx:
+            return float("nan") if not return_limits else (float("nan"), limits)
+        m = np.mean(scores[bottom_indx : top_indx + 1])
+        return m if not return_limits else (m, (scores[bottom_indx], scores[top_indx]))
+    elif base == "scores":
+        bottom_limit = min(1, max(0, bottom_limit))
+        top_limit = min(1, max(0, top_limit))
+        limits = (
+            np.quantile(scores, bottom_limit),
+            np.quantile(scores, 1 - top_limit),
+        )
+        if limits[0] > limits[1]:
+            return float("nan") if not return_limits else (float("nan"), limits)
+    elif base == "mean":
+        return np.mean(scores) if not return_limits else (np.mean(scores), None)
+    elif base == "median":
+        return np.median(scores) if not return_limits else (np.median(scores), None)
+    else:
+        raise ValueError(f"Unknown base for truncated_mean ({base})")
+    if len(scores) == 0 or limits[1] < limits[0]:
+        return float("nan") if not return_limits else (float("nan"), limits)
+    try:
+        # this is an inclusive trimmed mean
+        # tm = tmean(scores, limits)
+        scores = scores[scores >= limits[0]]
+        scores = scores[scores <= limits[1]]
+        if len(scores) == 0:
+            return float("nan") if not return_limits else (float("nan"), limits)
+        tm = np.mean(scores)
+        return tm if not return_limits else (tm, limits)
+    except ValueError:
+        return float("nan") if not return_limits else (float("nan"), limits)
+
+
+SINGLE_THREAD_FORCED = False
+
+
+def force_single_thread(on: bool = True):
+    """
+    Forces negmas to use a single thread for all internal calls.
+
+    Remarks:
+        - This will have the effect of not enforcing time-limits on calls.
+        - Only use this with caution and for debugging.
+    """
+    global SINGLE_THREAD_FORCED
+    SINGLE_THREAD_FORCED = on
+
+
+@contextmanager
+def single_thread():
+    force_single_thread(True)
+    yield None
+    force_single_thread(False)
+
+
+def is_single_thread() -> bool:
+    return SINGLE_THREAD_FORCED
+
+
+class TimeoutCaller:
+    pool = None
+
+    @classmethod
+    def run(cls, to_run, timeout: float):
+        if is_single_thread():
+            return to_run()
+        pool = cls.get_pool()
+        future = pool.submit(to_run)
+        try:
+            result = future.result(timeout)
+            return result
+        except TimeoutError as s:
+            future.cancel()
+            raise s
+
+    @classmethod
+    def get_pool(cls):
+        if cls.pool is None:
+            cls.pool = ThreadPoolExecutor()
+        return cls.pool
+
+    @classmethod
+    def cleanup(cls):
+        if cls.pool is not None:
+            try:
+                cls.pool.shutdown(wait=False)
+                for thread in cls.pool._threads:
+                    del concurrent.futures.thread._threads_queues[thread]
+            except:
+                warnings.warn(
+                    "NegMAS have finished processing but there are some "
+                    "threads still hanging there!! If your program does "
+                    "not die by itself. Please press Ctrl-c to kill it"
+                )
+
+
+atexit.register(TimeoutCaller.cleanup)
