@@ -3,18 +3,19 @@
 import functools
 import operator
 import os
+import pathlib
 import shutil
 import xml.etree.ElementTree as ET
 from os import listdir
-from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union
+from typing import Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union
 
 import numpy as np
-import pkg_resources
 
 from .generics import ivalues
+from .helpers import PATH
 from .negotiators import Negotiator
 from .outcomes import Issue, enumerate_outcomes
-from .sao import AspirationNegotiator, SAOMechanism
+from .sao import SAOMechanism
 from .utilities import UtilityFunction, make_discounted_ufun
 
 __all__ = [
@@ -30,7 +31,7 @@ __all__ = [
 def get_domain_issues(
     domain_file_name: str,
     force_single_issue=False,
-    max_n_outcomes: int = 1e6,
+    max_n_outcomes: int = 1_000_000,
     n_discretization: Optional[int] = None,
     keep_issue_names=True,
     keep_value_names=True,
@@ -52,7 +53,7 @@ def get_domain_issues(
         List or Dict of issues
 
     """
-    issues, issues_details, mechanism = None, None, None
+    issues, issues_details = None, None
     if domain_file_name is not None:
         domain_file_name = str(domain_file_name)
         issues_details, _ = Issue.from_genius(
@@ -87,7 +88,7 @@ def get_domain_issues(
 
 
 def load_genius_domain(
-    domain_file_name: str,
+    domain_file_name: PATH,
     utility_file_names: Optional[List[str]] = None,
     agent_factories: Optional[
         Union[Callable[[], Negotiator], List[Callable[[], Negotiator]]]
@@ -95,11 +96,12 @@ def load_genius_domain(
     force_single_issue=False,
     force_numeric=False,
     cache_and_discretize_outcomes=False,
-    max_n_outcomes: int = 1e6,
+    max_n_outcomes: int = 1_000_000,
     n_discretization: Optional[int] = None,
     keep_issue_names=True,
     keep_value_names=True,
-    normalize_utilities=True,
+    normalize_utilities=False,
+    normalize_max_only=False,
     n_steps=None,
     time_limit=3 * 60,  # GENIUS uses 3min time limit by default
     max_n_agents=None,
@@ -107,6 +109,7 @@ def load_genius_domain(
     safe_parsing=False,
     ignore_reserved=False,
     ignore_discount=False,
+    **kwargs,
 ) -> Tuple[Optional[SAOMechanism], List[dict], Union[Dict[str, Issue], List[Issue]]]:
     """
     Loads a genius domain, creates appropriate negotiators if necessary
@@ -122,6 +125,7 @@ def load_genius_domain(
         keep_issue_names:
         keep_value_names:
         normalize_utilities:
+        normalize_max_only:
         n_steps:
         time_limit:
         max_n_agents:
@@ -129,6 +133,7 @@ def load_genius_domain(
         safe_parsing:
         ignore_reserved:
         ignore_discount:
+        kwargs: Extra arguments to pass verbatim to SAOMechanism constructor
 
     Returns:
         - mechanism (SAOMechanism): A mechanism for the given issues
@@ -136,6 +141,7 @@ def load_genius_domain(
         - issues Union[Issue, Dict[str, Issue], List[Issue]]] : The issues
 
     """
+
     issues, issues_details, mechanism = None, None, None
     if domain_file_name is not None:
         domain_file_name = str(domain_file_name)
@@ -181,6 +187,7 @@ def load_genius_domain(
             keep_issue_names=keep_issue_names,
             keep_value_names=keep_value_names,
             normalize_utility=normalize_utilities,
+            normalize_max_only=normalize_max_only,
             domain_issues=issues_details,
             safe_parsing=safe_parsing,
             max_n_outcomes=max_n_outcomes,
@@ -222,10 +229,12 @@ def load_genius_domain(
     if domain_file_name is not None:
         mechanism_name = (
             domain_file_name.split("/")[-1][:-4]
-            .replace("-domain", "")
-            .replace("_domain", "")
-            .replace("domain", "")
+            # .replace("-domain", "")
+            # .replace("_domain", "")
+            # .replace("domain", "")
         )
+        kwargs["outcome_type"] = dict if keep_issue_names else tuple
+        kwargs["avoid_ultimatum"] = False
         mechanism = SAOMechanism(
             issues=issues,
             outcomes=outcomes,
@@ -234,8 +243,7 @@ def load_genius_domain(
             max_n_agents=max_n_agents,
             dynamic_entry=dynamic_entry,
             name=mechanism_name,
-            outcome_type=dict if keep_issue_names else tuple,
-            enable_callbacks=True,
+            **kwargs,
         )
         if agent_info is not None and len(agent_info) > 0:
             for info in agent_info:
@@ -268,18 +276,20 @@ def load_genius_domain_from_folder(
     force_single_issue=False,
     force_numeric=False,
     cache_and_discretize_outcomes=False,
-    max_n_outcomes: int = 1e6,
+    max_n_outcomes: int = 1_000_000,
     n_discretization: Optional[int] = None,
     keep_issue_names=True,
     keep_value_names=True,
-    normalize_utilities=True,
+    normalize_utilities=False,
+    normalize_max_only=False,
     n_steps=None,
-    time_limit=60,  # GENIUS uses 3min time limit by default
+    time_limit=3 * 60,  # GENIUS uses 3min time limit by default
     max_n_agents=None,
     dynamic_entry=True,
     safe_parsing=False,
     ignore_reserved=False,
     ignore_discount=False,
+    **kwargs,
 ) -> Tuple[Optional[SAOMechanism], List[dict], Union[Dict[str, Issue], List[Issue]]]:
     """
     Loads a genius domain from a folder. See ``load_genius_domain`` for more details.
@@ -294,6 +304,7 @@ def load_genius_domain_from_folder(
         keep_issue_names:
         keep_value_names:
         normalize_utilities:
+        normalize_max_only:
         n_steps:
         time_limit:
         max_n_agents:
@@ -301,14 +312,17 @@ def load_genius_domain_from_folder(
         safe_parsing:
         ignore_reserved:
         ignore_discount:
+        kwargs: other arguments to pass to SAOMechanism constructor verbatim
 
     Returns:
         - mechanism (SAOMechanism): A mechanism for the given issues
-        - agent_info (List[Dict]): All Negotiator functions from the given file
+        - agent_info (List[Dict]): All Negotiator utility functions from the given file
         - issues Union[Issue, Dict[str, Issue], List[Issue]]] : The issues
 
     Examples:
 
+        >>> import pkg_resources
+        >>> from negmas import *
         >>> folder_name = pkg_resources.resource_filename('negmas', resource_name='tests/data/10issues')
         >>> mechanism, negotiators, issues = load_genius_domain_from_folder(folder_name
         ...                             , force_single_issue=False, keep_issue_names=False
@@ -318,7 +332,7 @@ def load_genius_domain_from_folder(
         >>> print(len(negotiators))
         2
         >>> print([type(a['ufun']) for a in negotiators])
-        [<class 'negmas.utilities.HyperRectangleUtilityFunction'>, <class 'negmas.utilities.HyperRectangleUtilityFunction'>]
+        [<class 'negmas.utilities.nonlinear.HyperRectangleUtilityFunction'>, <class 'negmas.utilities.nonlinear.HyperRectangleUtilityFunction'>]
         >>> print(negotiators[0]['ufun'].outcome_ranges[0])
         {1: (7.0, 9.0), 3: (2.0, 7.0), 5: (0.0, 8.0), 8: (0.0, 7.0)}
         >>> print(negotiators[0]['ufun'].mappings[0])
@@ -347,7 +361,7 @@ def load_genius_domain_from_folder(
         (3, 2)
 
         >>> [type(a['ufun']) for a in negotiators]
-        [<class 'negmas.utilities.LinearUtilityAggregationFunction'>, <class 'negmas.utilities.LinearUtilityAggregationFunction'>]
+        [<class 'negmas.utilities.linear.LinearUtilityAggregationFunction'>, <class 'negmas.utilities.linear.LinearUtilityAggregationFunction'>]
 
         >>> mechanism, negotiators, issues = load_genius_domain_from_folder(
         ...                             pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
@@ -356,7 +370,7 @@ def load_genius_domain_from_folder(
         >>> len(issues), len(negotiators)
         (1, 2)
         >>> [type(a['ufun']) for a in negotiators]
-        [<class 'negmas.utilities.MappingUtilityFunction'>, <class 'negmas.utilities.MappingUtilityFunction'>]
+        [<class 'negmas.utilities.nonlinear.MappingUtilityFunction'>, <class 'negmas.utilities.nonlinear.MappingUtilityFunction'>]
 
     """
     folder_name = str(folder_name)
@@ -373,6 +387,8 @@ def load_genius_domain_from_folder(
             domain_file_name = full_name
         elif root.tag == "utility_space":
             utility_file_names.append(full_name)
+    if domain_file_name is None:
+        return None, [], []
     return load_genius_domain(
         domain_file_name=domain_file_name,
         utility_file_names=utility_file_names,
@@ -385,6 +401,7 @@ def load_genius_domain_from_folder(
         keep_issue_names=keep_issue_names,
         keep_value_names=keep_value_names,
         normalize_utilities=normalize_utilities,
+        normalize_max_only=normalize_max_only,
         n_steps=n_steps,
         time_limit=time_limit,
         max_n_agents=max_n_agents,
@@ -392,15 +409,16 @@ def load_genius_domain_from_folder(
         safe_parsing=safe_parsing,
         ignore_reserved=ignore_reserved,
         ignore_discount=ignore_discount,
+        **kwargs,
     )
 
 
-def find_domain_and_utility_files(folder_name) -> Tuple[str, List[str]]:
-    """Finds the domain and utility_function files in a folder
-    """
+def find_domain_and_utility_files(folder_name) -> Tuple[Optional[PATH], List[PATH]]:
+    """Finds the domain and utility_function files in a folder"""
     files = sorted(listdir(folder_name))
     domain_file_name = None
     utility_file_names = []
+    folder_name = str(folder_name)
     for f in files:
         if not f.endswith(".xml") or f.endswith("pareto.xml"):
             continue
@@ -415,19 +433,21 @@ def find_domain_and_utility_files(folder_name) -> Tuple[str, List[str]]:
 
 
 def convert_genius_domain(
-    src_domain_file_name: str,
-    dst_domain_file_name: str,
-    src_utility_file_names: Optional[List[str]] = None,
-    dst_utility_file_names: Optional[List[str]] = None,
+    src_domain_file_name: Optional[PATH],
+    dst_domain_file_name: Optional[PATH],
+    src_utility_file_names: Sequence[PATH] = tuple(),
+    dst_utility_file_names: Sequence[PATH] = tuple(),
     force_single_issue=False,
     cache_and_discretize_outcomes=False,
-    max_n_outcomes: int = 1e6,
+    max_n_outcomes: int = 1_000_000,
     n_discretization: Optional[int] = None,
     keep_issue_names=True,
     keep_value_names=True,
-    normalize_utilities=True,
+    normalize_utilities=False,
+    normalize_max_only=False,
     safe_parsing=False,
 ) -> bool:
+    assert len(src_utility_file_names) == len(dst_utility_file_names)
     if (
         not force_single_issue
         and not cache_and_discretize_outcomes
@@ -436,12 +456,14 @@ def convert_genius_domain(
         and not normalize_utilities
     ):
         # no need to do anything, just copy
-        shutil.copy(src=src_domain_file_name, dst=dst_domain_file_name)
+        if src_domain_file_name and dst_domain_file_name:
+            shutil.copy(src=src_domain_file_name, dst=dst_domain_file_name)
         for src, dst in zip(src_utility_file_names, dst_utility_file_names):
             shutil.copy(src=src, dst=dst)
         return True
-    issues, issues_details, mechanism = None, None, None
-    if src_domain_file_name is not None:
+    issues, issues_details = None, None
+
+    if src_domain_file_name is not None and dst_domain_file_name is not None:
         issues_details, _ = Issue.from_genius(
             src_domain_file_name,
             force_single_issue=False,
@@ -463,12 +485,14 @@ def convert_genius_domain(
             issues = issues_details
         if issues is None:
             return False
+
         Issue.to_genius(
             issues=issues, file_name=dst_domain_file_name, enumerate_integer=True
         )
 
     if src_utility_file_names is None:
         src_utility_file_names = []
+
     for ufname, dstfname in zip(src_utility_file_names, dst_utility_file_names):
         utility, discount_factor = UtilityFunction.from_genius(
             file_name=ufname,
@@ -476,6 +500,7 @@ def convert_genius_domain(
             keep_issue_names=keep_issue_names,
             keep_value_names=keep_value_names,
             normalize_utility=normalize_utilities,
+            normalize_max_only=normalize_max_only,
             domain_issues=issues_details,
             safe_parsing=safe_parsing,
             max_n_outcomes=max_n_outcomes,
@@ -493,13 +518,12 @@ def convert_genius_domain(
 
 
 def convert_genius_domain_from_folder(
-    src_folder_name: str, dst_folder_name: str, **kwargs
+    src_folder_name: PATH,
+    dst_folder_name: PATH,
+    **kwargs,
 ) -> bool:
-    """Loads a genius domain from a folder. See ``load_genius_domain`` for more details.
-
-
-
-    """
+    """Loads a genius domain from a folder. See ``load_genius_domain`` for more details."""
+    src_folder_name = pathlib.Path(src_folder_name)
     os.makedirs(dst_folder_name, exist_ok=True)
     files = sorted(listdir(src_folder_name))
     domain_file_name = None
@@ -507,8 +531,8 @@ def convert_genius_domain_from_folder(
     for f in files:
         if not f.endswith(".xml") or f.endswith("pareto.xml"):
             continue
-        full_name = src_folder_name + "/" + f
-        root = ET.parse(full_name).getroot()
+        full_name = src_folder_name / f
+        root = ET.parse(str(full_name)).getroot()
 
         if root.tag == "negotiation_template":
             domain_file_name = full_name
@@ -524,7 +548,7 @@ def convert_genius_domain_from_folder(
             os.path.join(dst_folder_name, os.path.basename(_))
             for _ in utility_file_names
         ],
-        **kwargs
+        **kwargs,
     )
 
     if not success:
