@@ -1,4 +1,5 @@
-"""Provides interfaces for defining negotiation mechanisms.
+"""
+Provides interfaces for defining negotiation mechanisms.
 """
 import pprint
 import random
@@ -31,12 +32,14 @@ from negmas.common import (
 )
 from negmas.events import Event, EventSource
 from negmas.generics import ikeys
+from negmas.genius import DEFAULT_JAVA_PORT, get_free_tcp_port
 from negmas.helpers import snake_case
 from negmas.negotiators import Negotiator
 from negmas.outcomes import (
     Issue,
     Outcome,
     enumerate_outcomes,
+    outcome_as_dict,
     outcome_as_tuple,
     outcome_is_valid,
 )
@@ -72,6 +75,35 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
     Base class for all negotiation Mechanisms.
 
     Override the `round` function of this class to implement a round of your mechanism
+
+    Args:
+        issues: List of issues to use (optional as you can pass `outcomes`)
+        outcomes: List of outcomes (optional as you can pass `issues`). If an int then it is the number of outcomes
+        n_steps: Number of rounds allowed (None means infinity)
+        time_limit: Number of real seconds allowed (None means infinity)
+        max_n_agents:  Maximum allowed number of agents
+        dynamic_entry: Allow agents to enter/leave negotiations between rounds
+        cache_outcomes: If true, a list of all possible outcomes will be cached
+        max_n_outcomes: The maximum allowed number of outcomes in the cached set
+        keep_issue_names: DEPRICATED. Use `outcome_type` instead. If True, dicts with issue names will be used for outcomes otherwise tuples
+        annotation: Arbitrary annotation
+        state_factory: A callable that receives an arbitrary set of key-value pairs and return a MechanismState
+                      descendant object
+        checkpoint_every: The number of steps to checkpoint after. Set to <= 0 to disable
+        checkpoint_folder: The folder to save checkpoints into. Set to None to disable
+        checkpoint_filename: The base filename to use for checkpoints (multiple checkpoints will be prefixed with
+                             step number).
+        single_checkpoint: If true, only the most recent checkpoint will be saved.
+        extra_checkpoint_info: Any extra information to save with the checkpoint in the corresponding json file as
+                               a dictionary with string keys
+        exist_ok: IF true, checkpoints override existing checkpoints with the same filename.
+        name: Name of the mechanism session. Should be unique. If not given, it will be generated.
+        outcome_type: The type used for representing outcomes. Can be tuple, dict or any `OutcomeType`
+        genius_port: the port used to connect to Genius for all negotiators in this mechanism (0 means any).
+        id: An optional system-wide unique identifier. You should not change
+            the default value except in special circumstances like during
+            serialization and should always guarantee system-wide uniquness
+            if you set this value explicitly
     """
 
     def __init__(
@@ -81,6 +113,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         n_steps: int = None,
         time_limit: float = None,
         step_time_limit: float = None,
+        negotiator_time_limit: float = None,
         max_n_agents: int = None,
         dynamic_entry=False,
         cache_outcomes=True,
@@ -97,34 +130,10 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         exist_ok: bool = True,
         name=None,
         outcome_type=tuple,
+        genius_port: int = DEFAULT_JAVA_PORT,
+        id: str = None,
     ):
-        """
-
-        Args:
-            issues: List of issues to use (optional as you can pass `outcomes`)
-            outcomes: List of outcomes (optional as you can pass `issues`). If an int then it is the number of outcomes
-            n_steps: Number of rounds allowed (None means infinity)
-            time_limit: Number of real seconds allowed (None means infinity)
-            max_n_agents:  Maximum allowed number of agents
-            dynamic_entry: Allow agents to enter/leave negotiations between rounds
-            cache_outcomes: If true, a list of all possible outcomes will be cached
-            max_n_outcomes: The maximum allowed number of outcomes in the cached set
-            keep_issue_names: DEPRICATED. Use `outcome_type` instead. If True, dicts with issue names will be used for outcomes otherwise tuples
-            annotation: Arbitrary annotation
-            state_factory: A callable that receives an arbitrary set of key-value pairs and return a MechanismState
-                          descendant object
-            checkpoint_every: The number of steps to checkpoint after. Set to <= 0 to disable
-            checkpoint_folder: The folder to save checkpoints into. Set to None to disable
-            checkpoint_filename: The base filename to use for checkpoints (multiple checkpoints will be prefixed with
-                                 step number).
-            single_checkpoint: If true, only the most recent checkpoint will be saved.
-            extra_checkpoint_info: Any extra information to save with the checkpoint in the corresponding json file as
-                                   a dictionary with string keys
-            exist_ok: IF true, checkpoints override existing checkpoints with the same filename.
-            name: Name of the mechanism session. Should be unique. If not given, it will be generated.
-            outcome_type: The type used for representing outcomes. Can be tuple, dict or any `OutcomeType`
-        """
-        super().__init__(name=name)
+        super().__init__(name, id=id)
         CheckpointMixin.checkpoint_init(
             self,
             step_attrib="_step",
@@ -139,6 +148,10 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         step_time_limit = (
             step_time_limit if step_time_limit is not None else float("inf")
         )
+        negotiator_time_limit = (
+            negotiator_time_limit if negotiator_time_limit is not None else float("inf")
+        )
+
         if keep_issue_names is not None:
             warnings.warn(
                 "keep_issue_names is depricated. Use outcome_type instead.\n"
@@ -162,7 +175,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
                 issues = []
                 issue_names = ikeys(outcomes[0])
                 for issue in range(n_issues):
-                    vals = list(set([_[issue] for _ in outcomes]))
+                    vals = list({_[issue] for _ in outcomes})
                     issues.append(vals)
                 __issues = [
                     Issue(_, name=name_) for _, name_ in zip(issues, issue_names)
@@ -197,7 +210,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
                     f", type={type(outcomes[0])})"
                 )
                 if keep_issue_names:
-                    for i, outcome in enumerate(outcomes):
+                    for outcome in outcomes:
                         assert list(outcome.keys()) == issue_names
                 else:
                     for i, outcome in enumerate(outcomes):
@@ -219,6 +232,8 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             )
 
         self.id = str(uuid.uuid4())
+        _imap = dict(zip((_.name for _ in __issues), range(len(__issues))))
+        _imap.update(dict(zip(range(len(__issues)), (_.name for _ in __issues))))
         self.ami = AgentMechanismInterface(
             id=self.id,
             n_outcomes=None if outcomes is None else len(outcomes),
@@ -227,10 +242,12 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             time_limit=time_limit,
             n_steps=n_steps,
             step_time_limit=step_time_limit,
+            negotiator_time_limit=negotiator_time_limit,
             dynamic_entry=dynamic_entry,
             max_n_agents=max_n_agents,
             annotation=annotation,
             outcome_type=dict if keep_issue_names else tuple,
+            imap=_imap,
         )
         self.ami._mechanism = self
 
@@ -248,6 +265,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         self._requirements = {}
         self._negotiators = []
         self._negotiator_map: Dict[str, "SAONegotiator"] = dict()
+        self._negotiator_index: Dict[str, int] = dict()
         self._roles = []
         self._start_time = None
         self._started = False
@@ -265,6 +283,18 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
 
         self.agents_of_role = defaultdict(list)
         self.role_of_agent = {}
+        # mechanisms do not differentiate between RANDOM_JAVA_PORT and ANY_JAVA_PORT.
+        # if either is given as the genius_port, it will fix a port and all negotiators
+        # that are not explicitly assigned to a port (by passing port>0 to them) will just
+        # use that port.
+        self.genius_port = genius_port if genius_port > 0 else get_free_tcp_port()
+
+        self.params = dict(
+            dynamic_entry=dynamic_entry,
+            genius_port=genius_port,
+            cache_outcomes=cache_outcomes,
+            annotation=annotation,
+        )
 
     def outcome_index(self, outcome) -> Optional[int]:
         """Returns the index of the outcome if that was possible"""
@@ -385,7 +415,9 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             return None
 
         relative_step = (
-            self._step / self.ami.n_steps if self.ami.n_steps is not None else -1.0
+            (self._step + 1) / (self.ami.n_steps + 1)
+            if self.ami.n_steps is not None
+            else -1.0
         )
         relative_time = (
             self.time / self.ami.time_limit if self.ami.time_limit is not None else -1.0
@@ -450,9 +482,13 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         ):
             self._negotiators.append(negotiator)
             self._negotiator_map[negotiator.id] = negotiator
+            self._negotiator_index[negotiator.id] = len(self._negotiators) - 1
             self._roles.append(role)
             self.role_of_agent[negotiator.uuid] = role
             self.agents_of_role[role].append(negotiator)
+            if negotiator.ufun is not None:
+                if hasattr(negotiator.ufun, "ami"):
+                    negotiator.ufun.ami = self.ami
             return True
         return None
 
@@ -478,6 +514,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             return False
         self._negotiators.remove(negotiator)
         self._negotiator_map.pop(negotiator.id)
+        self._negotiator_index.pop(negotiator.id)
         if self._enable_callbacks:
             negotiator.on_leave(self.ami, **kwargs)
         return True
@@ -502,6 +539,22 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
     def negotiators(self):
         return self._negotiators
 
+    @property
+    def negotiator_ids(self) -> List[str]:
+        return [_.id for _ in self._negotiators]
+
+    @property
+    def agent_ids(self) -> List[str]:
+        return [_.owner.id for _ in self._negotiators if _.owner]
+
+    @property
+    def agent_names(self) -> List[str]:
+        return [_.owner.name for _ in self._negotiators if _.owner]
+
+    @property
+    def negotiator_names(self) -> List[str]:
+        return [_.name for _ in self._negotiators]
+
     @property
     def requirements(self):
         """A dictionary specifying the requirements that must be in the capabilities of any agent
@@ -713,7 +766,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         if self._enable_callbacks:
             for agent in self._negotiators:
                 agent.on_round_end(state)
-        self._history.append(state4history)
+        self._add_to_history(state4history)
         self._step += 1
         self.on_negotiation_end()
         return state
@@ -741,67 +794,83 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         self.checkpoint_on_step_started()
         state = self.state
         state4history = self.state4history
-        if self.time > self.time_limit:
-            self._agreement, self._broken, self._timedout = None, False, True
-            self._history.append(state4history)
+
+        # end with a timeout if condition is met
+        if (self.time > self.time_limit) or (
+            self.ami.n_steps and self._step >= self.ami.n_steps
+        ):
+            self._running, self._broken, self._timedout = False, False, True
             self.on_negotiation_end()
-            return state
+            return self.state
+
+        # if there is a single negotiator and no other negotiators can be added,
+        # end without starting
         if len(self._negotiators) < 2:
             if self.ami.dynamic_entry:
-                self._history.append(state4history)
-                return state
+                return self.state
             else:
-                self.ami.state.running = False
-                self._agreement, self._broken, self._timedout = None, False, False
-                self._history.append(state4history)
+                self._running, self._broken, self._timedout = False, False, False
                 self.on_negotiation_end()
-                return state
+                return self.state
 
+        # if the mechanism states that it is broken, timedout or ended with
+        # agreement, report that
         if self._broken or self._timedout or self._agreement is not None:
-            self._history.append(state4history)
-            return state
+            self._running = False
+            self.on_negotiation_end()
+            return self.state
 
         if not self._running:
+            # if we did not start, just start
             self._running = True
             self._step = 0
             self._start_time = time.perf_counter()
             self._started = True
+            state = self.state
+            # if the mechanism indicates that it cannot start, keep trying
             if self.on_negotiation_start() is False:
                 self._agreement, self._broken, self._timedout = None, False, False
-                self._history.append(state4history)
-                return state
+                return self.state
             for a in self.negotiators:
                 a.on_negotiation_start(state=state)
             self.announce(Event(type="negotiation_start", data=None))
         else:
+            # if no steps are remaining, end with a timeout
             remaining_steps, remaining_time = self.remaining_steps, self.remaining_time
             if (remaining_steps is not None and remaining_steps <= 0) or (
                 remaining_time is not None and remaining_time <= 0.0
             ):
                 self._running = False
                 self._agreement, self._broken, self._timedout = None, False, True
-                self._history.append(state4history)
                 self.on_negotiation_end()
-                return state
+                return self.state
 
-        if not self._waiting:
-            if self._enable_callbacks:
-                for agent in self._negotiators:
-                    agent.on_round_start(state)
+        # send round start only if the mechanism is not waiting for anyone
+        # TODO check this.
+        if not self._waiting and self._enable_callbacks:
+            for agent in self._negotiators:
+                agent.on_round_start(state)
+
+        # run a round of the mechanism and get the new state
+        step_start = time.perf_counter() if not self._waiting else self._last_start
+        self._last_start = step_start
         self._waiting = False
-        step_start = time.perf_counter()
         result = self.round()
         step_time = time.perf_counter() - step_start
         self._stats["round_times"].append(step_time)
+
+        # if negotaitor times are reported, save them
         if result.times:
             for k, v in result.times.items():
                 if v is not None:
                     self._stats["times"][k] += v
+        # if negotaitor exceptions are reported, save them
         if result.exceptions:
             for k, v in result.exceptions.items():
                 if v:
                     self._stats["exceptions"][k] += v
-        state = self.state
+
+        # update current state variables from the result of the round just run
         self._error, self._error_details, self._waiting = (
             result.error,
             result.error_details,
@@ -822,15 +891,30 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             )
         if (self._agreement is not None) or self._broken or self._timedout:
             self._running = False
+
+        # now switch to the new state
+        state = self.state
         if not self._waiting:
+            state4history = self.state4history
             if self._enable_callbacks:
                 for agent in self._negotiators:
                     agent.on_round_end(state)
-            self._history.append(state4history)
+            self._add_to_history(state4history)
+            # we only indicate a new step if no one is waiting
             self._step += 1
         if not self._running:
             self.on_negotiation_end()
-        return state
+        return self.state
+
+    def _add_to_history(self, state4history):
+        if len(self._history) == 0:
+            self._history.append(state4history)
+            return
+        last = self._history[-1]
+        if last["step"] == state4history:
+            self._history[-1] = state4history
+            return
+        self._history.append(state4history)
 
     @classmethod
     def runall(
@@ -851,13 +935,13 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         """
         completed = [_ is None for _ in mechanisms]
         states = [None] * len(mechanisms)
+        indices = list(range(len(list(mechanisms))))
         if method == "serial":
             while not all(completed):
-                lst = zip(completed, mechanisms)
                 if not keep_order:
-                    lst = list(lst)
-                    random.shuffle(lst)
-                for i, (done, mechanism) in enumerate(lst):
+                    random.shuffle(indices)
+                for i in indices:
+                    done, mechanism = completed[i], mechanisms[i]
                     if done:
                         continue
                     result = mechanism.step()
@@ -893,22 +977,20 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             - List of states of all mechanisms after completion
 
         """
+        indices = list(range(len(list(mechanisms))))
         if not keep_order:
-            raise NotImplementedError(
-                "running mechanisms in random order is not yet supported"
-            )
+            random.shuffle(indices)
 
         completed = [_ is None for _ in mechanisms]
-        states = [None] * len(mechanisms)
-        for i, (done, mechanism) in enumerate(zip(completed, mechanisms)):
+        for i in indices:
+            done, mechanism = completed[i], mechanisms[i]
             if done:
                 continue
             result = mechanism.step()
             if result.running:
                 continue
             completed[i] = True
-            states[i] = mechanism.state
-        return states
+        return [_.state for _ in mechanisms]
 
     def run(self, timeout=None) -> MechanismState:
         if timeout is None:
@@ -918,7 +1000,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             start_time = time.perf_counter()
             for _ in self:
                 if time.perf_counter() - start_time > timeout:
-                    self._running, self._timedout = False, True
+                    self._running, self._timedout, self._broken = False, True, False
                     self.on_negotiation_end()
                     break
         return self.state
@@ -976,7 +1058,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
     @property
     def state(self):
         """Returns the current state. Override `extra_state` if you want to keep extra state"""
-        return self._state_factory(
+        d = dict(
             running=self._running,
             step=self._step,
             time=self.time,
@@ -989,7 +1071,12 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             has_error=self._error,
             error_details=self._error_details,
             waiting=self._waiting,
-            **self.extra_state(),
+        )
+        d2 = self.extra_state()
+        if d2:
+            d.update(d2)
+        return self._state_factory(
+            **d,
         )
 
     def pareto_frontier(
@@ -1023,7 +1110,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
 
     @abstractmethod
     def round(self) -> MechanismRoundResult:
-        """ Implements a single step of the mechanism. Override this!
+        """Implements a single step of the mechanism. Override this!
 
         Returns:
             MechanismRoundResult giving whether the negotiation was broken or timedout and the agreement if any.
@@ -1040,6 +1127,14 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
     def _get_ami(self, negotiator: Negotiator, role: str) -> AgentMechanismInterface:
         return self.ami
 
+    def cast_outcome(self, outcome):
+        """Converts an outcome to the outcome-type used in this negotiation"""
+        if issubclass(self.ami.outcome_type, tuple) and not isinstance(outcome, tuple):
+            return outcome_as_tuple(outcome)
+        if issubclass(self.ami.outcome_type, dict) and not isinstance(outcome, tuple):
+            return outcome_as_dict(outcome, self.issues)
+        return outcome
+
 
 # @dataclass
 # class MechanismSequenceState(MechanismState):
