@@ -5,7 +5,6 @@ from dataclasses import dataclass, field
 from typing import List, Optional, Tuple
 
 from .mechanisms import Mechanism, MechanismRoundResult, MechanismState
-from .outcomes import Outcome
 
 __all__ = [
     "VetoMTMechanism",
@@ -71,7 +70,9 @@ class VetoMTMechanism(Mechanism):
         self.epsilon = epsilon
 
     def extra_state(self):
-        return dict(current_offers=deepcopy(self.current_offers),)
+        return dict(
+            current_offers=deepcopy(self.current_offers),
+        )
 
     def next_outcome(self, outcome: Optional["Outcome"]) -> Optional["Outcome"]:
         """Generate the next outcome given some outcome.
