@@ -7,6 +7,7 @@ import functools
 import math
 import warnings
 from abc import ABC
+from collections import namedtuple
 from random import sample
 from typing import (
     TYPE_CHECKING,
@@ -29,13 +30,14 @@ from negmas.outcomes import Issue
 
 if TYPE_CHECKING:
     from negmas.outcomes import Outcome
-    from negmas.utilities import UtilityValue, UtilityFunction
     from negmas.situated import Agent
+    from negmas.utilities import UtilityFunction
 
 __all__ = [
     "Negotiator",
     "AspirationMixin",
     "Controller",
+    "NegotiatorInfo",
     "PassThroughNegotiator",
     "EvaluatorMixin",
     "RealComparatorMixin",
@@ -53,18 +55,21 @@ __all__ = [
     "SorterNegotiator",
 ]
 
+NegotiatorInfo = namedtuple("NegotiatorInfo", ["negotiator", "context"])
+"""The return type of `negotiators` member of `Controller`."""
+
 
 class Negotiator(Rational, Notifiable, ABC):
     r"""Abstract negotiation agent. Base class for all negotiators
 
-     Args:
+    Args:
 
-            name: Negotiator name. If not given it is assigned by the system (unique 16 characters).
+           name: Negotiator name. If not given it is assigned by the system (unique 16 characters).
 
-        Returns:
-            bool: True if participating in the given negotiation (or any negotiation if it was None)
+       Returns:
+           bool: True if participating in the given negotiation (or any negotiation if it was None)
 
-        Remarks:
+       Remarks:
 
     """
 
@@ -74,8 +79,9 @@ class Negotiator(Rational, Notifiable, ABC):
         ufun: Optional["UtilityFunction"] = None,
         parent: "Controller" = None,
         owner: "Agent" = None,
+        id: str = None,
     ) -> None:
-        super().__init__(name=name, ufun=ufun)
+        super().__init__(name=name, ufun=ufun, id=id)
         self.__parent = parent
         self._capabilities = {"enter": True, "leave": True, "ultimatum": True}
         self._mechanism_id = None
@@ -338,7 +344,7 @@ class Negotiator(Rational, Notifiable, ABC):
               it.
         """
         if hasattr(self._utility_function, "outcome_type"):
-            if self._utility_function.outcome_type is None:
+            if self._ami and self._utility_function.outcome_type is None:
                 self._utility_function.outcome_type = self._ami.outcome_type
                 self._utility_function.issue_names = [_.name for _ in self._ami.issues]
             elif (
@@ -351,6 +357,15 @@ class Negotiator(Rational, Notifiable, ABC):
                 )
         super().on_ufun_changed()
 
+    def cancel(self, reason=None) -> None:
+        """
+        A method that may be called by a mechanism to make the negotiator cancel whatever it is currently
+        processing.
+
+        Negotiators can just ignore this message (default behavior) but if there is a way to actually cancel
+        work, it should be implemented here to improve the responsiveness of the negotiator.
+        """
+
     def __str__(self):
         return f"{self.name}"
 
@@ -372,14 +387,18 @@ class PassThroughNegotiator(Negotiator):
     """
 
     def __getattribute__(self, item):
-        if item in (
-            "id",
-            "name",
-            "on_ufun_changed",
-            "has_ufun",
-            "utility_function",
-            "reserved_value",
-        ) or item.startswith("_"):
+        if (
+            item
+            in (
+                "id",
+                "name",
+                "on_ufun_changed",
+                "has_ufun",
+                "utility_function",
+                "reserved_value",
+            )
+            or item.startswith("_")
+        ):
             return super().__getattribute__(item)
         parent = super().__getattribute__("__dict__").get("_Negotiator__parent", None)
         if parent is None:
@@ -433,11 +452,10 @@ class Controller(Rational):
         default_negotiator_params: Dict[str, Any] = None,
         parent: Union["Controller", "Agent"] = None,
         auto_kill: bool = True,
-        name: str = None,
-        ufun: "UtilityFunction" = None,
+        **kwargs,
     ):
-        super().__init__(name=name, ufun=ufun)
-        self._negotiators: Dict[str, Tuple["PassThroughNegotiator", Any]] = {}
+        super().__init__(**kwargs)
+        self._negotiators: Dict[str, NegotiatorInfo] = {}
         if default_negotiator_params is None:
             default_negotiator_params = {}
         if isinstance(default_negotiator_type, str):
@@ -448,7 +466,7 @@ class Controller(Rational):
         self._auto_kill = auto_kill
 
     @property
-    def negotiators(self) -> Dict[str, Tuple[Negotiator, Any]]:
+    def negotiators(self) -> Dict[str, NegotiatorInfo]:
         """
         Returns a dictionary mapping negotiator ID to the a tuple containing
         the negotiator and its context
@@ -456,15 +474,17 @@ class Controller(Rational):
         return self._negotiators
 
     @property
-    def active_negotiators(self) -> Dict[str, Tuple[Negotiator, Any]]:
+    def active_negotiators(self) -> Dict[str, NegotiatorInfo]:
         """
         Returns the negotiators whose negotiations are running.
         Returns a dictionary mapping negotiator ID to the a tuple containing the negotiator
-        and its context"""
+        and its context
+        """
         return {
             k: v
             for k, v in self._negotiators.items()
-            if v[0].ami is not None and v[0].ami.state.running
+            if v[0].ami is not None
+            and (v[0].ami.state.running or not v[0].ami.state.started)
         }
 
     @property
@@ -495,7 +515,31 @@ class Controller(Rational):
 
         Returns:
 
-            PassThroughNegotiator: The negotiator to be controlled
+            The negotiator to be controlled. None for failure
+
+        """
+        new_negotiator = self.make_negotiator(negotiator_type, name, **kwargs)
+        self.add_negotiator(new_negotiator)
+        return new_negotiator
+
+    def make_negotiator(
+        self,
+        negotiator_type: Union[str, Type[PassThroughNegotiator]] = None,
+        name: str = None,
+        **kwargs,
+    ) -> PassThroughNegotiator:
+        """
+        Creates a negotiator but does not add it to the controller. Call
+        `add_negotiator` to add it.
+
+        Args:
+            negotiator_type: Type of the negotiator to be created
+            name: negotiator name
+            **kwargs: any key-value pairs to be passed to the negotiator constructor
+
+        Returns:
+
+            The negotiator to be controlled. None for failure
 
         """
         if negotiator_type is None:
@@ -510,15 +554,31 @@ class Controller(Rational):
         args = self.__default_negotiator_params
         if kwargs:
             args.update(kwargs)
-        new_negotiator = negotiator_type(name=name, parent=self, **args)
-        if new_negotiator is not None:
-            self._negotiators[new_negotiator.id] = (new_negotiator, cntxt)
-        return new_negotiator
+        return negotiator_type(name=name, parent=self, **args)
+
+    def add_negotiator(
+        self,
+        negotiator: Negotiator,
+        cntxt: Any = None,
+    ) -> None:
+        """
+        Adds a negotiator to the controller
+
+        Args:
+            negotaitor: The negotaitor to add
+            name: negotiator name
+            cntxt: The context to be associated with this negotiator.
+            **kwargs: any key-value pairs to be passed to the negotiator constructor
+
+        """
+        if negotiator is not None:
+            self._negotiators[negotiator.id] = NegotiatorInfo(negotiator, cntxt)
 
     def call(self, negotiator: PassThroughNegotiator, method: str, *args, **kwargs):
         """
-        Calls the given method on the given negotiator safely without causing recursion. The controller MUST use this
-        function to access any callable on the negotiator
+        Calls the given method on the given negotiator safely without causing
+        recursion. The controller MUST use this function to access any callable
+        on the negotiator.
 
         Args:
             negotiator:
@@ -556,6 +616,60 @@ class Controller(Rational):
             negotiator._Negotiator__parent = None
             self._negotiators.pop(negotiator_id, None)
 
+    def partner_negotiator_ids(self, negotiator_id: str) -> Optional[List[str]]:
+        """
+        Finds the negotiator ID negotiating with one of our negotiators.
+
+        Args:
+
+            negotiator_id: Our negotiator ID
+        """
+        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
+        if not negotiator or not negotiator.ami:
+            return None
+        return [_ for _ in negotiator.ami.negotiator_ids if _ != negotiator_id]
+
+    def partner_negotiator_names(self, negotiator_id: str) -> Optional[List[str]]:
+        """
+        Finds the negotiator names negotiating with one of our negotiators.
+
+        Args:
+
+            negotiator_id: Our negotiator ID
+        """
+        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
+        if not negotiator or not negotiator.ami:
+            return None
+        return [_ for _ in negotiator.ami.negotiator_names if _ != negotiator.name]
+
+    def partner_agent_ids(self, negotiator_id: str) -> Optional[List[str]]:
+        """
+        Finds the agent ID negotiating with one of our negotiators.
+
+        Args:
+
+            negotiator_id: Our negotiator ID
+        """
+        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
+        if not negotiator or not negotiator.ami:
+            return None
+        me = negotiator.owner.id if negotiator.owner else ""
+        return [_ for _ in negotiator.ami.agent_ids if _ and _ != me]
+
+    def partner_agent_names(self, negotiator_id: str) -> Optional[List[str]]:
+        """
+        Finds the negotiator names negotiating with one of our negotiators.
+
+        Args:
+
+            negotiator_id: Our negotiator ID
+        """
+        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
+        if not negotiator or not negotiator.ami:
+            return None
+        me = negotiator.owner.name if negotiator.owner else ""
+        return [_ for _ in negotiator.ami.agent_names if _ and _ != me]
+
     def before_join(
         self,
         negotiator_id: str,
@@ -627,7 +741,7 @@ class Controller(Rational):
             negotiation.
 
         """
-        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
+        negotiator, _ = self._negotiators.get(negotiator_id, (None, None))
         if negotiator is None:
             raise ValueError(f"Unknown negotiator {negotiator_id}")
         permission = self.before_join(negotiator, ami, state, ufun=ufun, role=role)
@@ -690,7 +804,7 @@ class Controller(Rational):
             negotiator_id: The negotiator ID
             state: `MechanismState` giving current state of the negotiation.
 
-       """
+        """
         negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
         if negotiator is None:
             raise ValueError(f"Unknown negotiator {negotiator_id}")
@@ -899,8 +1013,8 @@ class NLevelsComparatorMixin:
             if isinstance(scale, str):
                 scale = dict(
                     linear=lambda x: x,
-                    log=lambda x: math.log(x),
-                    exp=lambda x: math.exp(x),
+                    log=math.log,
+                    exp=math.exp,
                 ).get(scale, None)
                 if scale is None:
                     raise ValueError(f"Unknown scale function {scale}")
@@ -935,7 +1049,7 @@ class NLevelsComparatorMixin:
             for second in sample(samples[i + 1 :], k=n_diffs):
                 diffs.append(abs(ufun.compare_real(first, second)))
         diffs = np.array(diffs)
-        hist, edges = np.histogram(diffs, bins=n + 1)
+        _, edges = np.histogram(diffs, bins=n + 1)
         return edges[1:-1].tolist()
 
     @property
@@ -1063,7 +1177,7 @@ class RankerWithWeightsMixin:
 
 class RankerMixin:
     """Adds the ability to rank outcomes returning the ranks without weights. Outcomes of equal utility are ordered
-     arbitrarily. None stands for the null outcome"""
+    arbitrarily. None stands for the null outcome"""
 
     def init(self):
         self.capabilities["rank"] = True
@@ -1105,7 +1219,7 @@ class RankerMixin:
 
 class SorterMixin:
     """Adds the ability to sort outcomes according to utility. Outcomes of equal utility are ordered
-     arbitrarily. None stands for the null outcome"""
+    arbitrarily. None stands for the null outcome"""
 
     def init(self):
         self.capabilities["sort"] = True
@@ -1184,11 +1298,11 @@ class NLevelsComparatorNegotiator(NLevelsComparatorMixin, Negotiator):
 class RankerWithWeightsNegotiator(RankerWithWeightsMixin, Negotiator):
     """A negotiator that can be asked to rank outcomes returning rank and weight. By default is just consults the ufun.
 
-        To change that behavior, override `rank_with_weights`.
+    To change that behavior, override `rank_with_weights`.
 
-        It has the `rank-weighted` capability.
+    It has the `rank-weighted` capability.
 
-        """
+    """
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
