@@ -1,4 +1,5 @@
-"""Defines basic concept related to outcomes
+"""
+Defines basic concepts related to outcomes
 
 Outcomes in this package are always assumed to be multi-issue outcomes where  single-issue outcomes can be implemented
 as the special case with a single issue.
@@ -28,6 +29,7 @@ import copy
 import itertools
 import numbers
 import random
+import sys
 import warnings
 import xml.etree.ElementTree as ET
 from collections import defaultdict
@@ -45,7 +47,6 @@ from typing import (
     Iterable,
     List,
     Mapping,
-    NewType,
     Optional,
     Sequence,
     Tuple,
@@ -57,7 +58,7 @@ import numpy as np
 
 from .common import NamedObject
 from .generics import ienumerate, iget, ikeys, ivalues
-from .helpers import unique_name
+from .helpers import PATH, unique_name
 from .java import PYTHON_CLASS_IDENTIFIER
 
 if TYPE_CHECKING:
@@ -84,6 +85,8 @@ __all__ = [
     "is_outcome",
     "num_outcomes",
     "outcome_for",
+    "outcome_types_are_ok",
+    "cast_outcome",
 ]
 
 
@@ -160,8 +163,9 @@ class Issue(NamedObject):
         min_value: Any = None,
         max_value: Any = None,
         value_type: Optional[Type] = None,
+        id=None,
     ) -> None:
-        super().__init__(name=name)
+        super().__init__(name, id=id)
 
         if (
             value_type is not None
@@ -173,8 +177,8 @@ class Issue(NamedObject):
             )
         # if isinstance(values, int) and values <= LARGE_NUMBER:
         #    values = list(range(values))
-        self._value_type: Type = value_type
-        self._n_values: int = float("inf")
+        self._value_type: Optional[Type] = value_type
+        self._n_values: Union[int, float] = float("inf")
         self._is_generator: bool = False
         self._is_range: bool = False
         self._is_int_range: bool = False
@@ -254,12 +258,16 @@ class Issue(NamedObject):
         self._values = values
 
         if isinstance(self._values, tuple):
-            self.min_value, self.max_value = values
+            self.min_value, self.max_value = self._values
         elif isinstance(self._values, list):
-            self.min_value, self.max_value = min(values), max(values)
+            self.min_value, self.max_value = min(self._values), max(self._values)
         else:
             self.min_value, self.max_value = min_value, max_value
 
+    @property
+    def value_type(self):
+        return self._value_type
+
     @property
     def values(self):
         return self._values
@@ -459,7 +467,7 @@ class Issue(NamedObject):
 
     @classmethod
     def to_genius(
-        cls, issues: Iterable["Issue"], file_name: str, enumerate_integer: bool = True
+        cls, issues: Iterable["Issue"], file_name: PATH, enumerate_integer: bool = True
     ) -> None:
         """Exports a the domain issues to a GENIUS XML file.
 
@@ -525,8 +533,8 @@ class Issue(NamedObject):
         keep_issue_names=True,
         safe_parsing=True,
         n_discretization: Optional[int] = None,
-        max_n_outcomes: int = 1e6,
-    ):
+        max_n_outcomes: int = 1_000_000,
+    ) -> Tuple[Optional[List["Issue"]], Optional[List[str]]]:
         """Exports a list/dict of issues from a GENIUS XML file.
 
         Args:
@@ -655,16 +663,15 @@ class Issue(NamedObject):
             if safe_parsing:
                 raise ValueError(f"No objective child was found in the root")
             utility_space = root
-        weights = {}
-        issues = {}
+        issues_dict: Dict[Union[int, str], Any] = {}
         issue_info = {}
         all_discrete = True
 
         for child in utility_space:
             if child.tag == "issue":
                 indx = int(child.attrib["index"]) - 1
-                myname = child.attrib["name"]
-                issue_key = myname if keep_issue_names else indx
+                myname = str(child.attrib["name"])
+                issue_key: Union[int, str] = myname if keep_issue_names else indx
                 issue_info[issue_key] = {"name": myname, "index": indx}
                 info = {"type": "discrete", "etype": "discrete", "vtype": "discrete"}
 
@@ -673,7 +680,7 @@ class Issue(NamedObject):
                 mytype = info["type"]
 
                 if mytype == "discrete":
-                    issues[issue_key] = []
+                    issues_dict[issue_key] = []
 
                     for item in child:
                         if item.tag == "item":
@@ -688,52 +695,71 @@ class Issue(NamedObject):
                             )
 
                             if (
-                                item_key not in issues[issue_key]
+                                item_key not in issues_dict[issue_key]
                             ):  # ignore repeated items
-                                issues[issue_key].append(item_key)
+                                issues_dict[issue_key].append(item_key)
 
                     if not keep_value_names:
-                        issues[issue_key] = len(issues[issue_key])
+                        issues_dict[issue_key] = len(issues_dict[issue_key])
                 elif mytype in ("integer", "real"):
-                    lower, upper = (
+                    lower_, upper_ = (
                         child.attrib.get("lowerbound", None),
                         child.attrib.get("upperbound", None),
                     )
 
                     for rng_child in child:
                         if rng_child.tag == "range":
-                            lower, upper = (
-                                rng_child.attrib.get("lowerbound", lower),
-                                rng_child.attrib.get("upperbound", upper),
+                            lower_, upper_ = (
+                                rng_child.attrib.get("lowerbound", lower_),
+                                rng_child.attrib.get("upperbound", upper_),
                             )
-
+                    if lower_ is None:
+                        if upper_ is not None and float(upper_) < 0:
+                            lower_ = str(
+                                -(sys.maxsize // 2)
+                                if mytype == "integer"
+                                else float("-inf")
+                            )
+                        else:
+                            lower_ = "0"
+                    if upper_ is None:
+                        upper_ = str(
+                            (sys.maxsize // 2) if mytype == "integer" else float("-inf")
+                        )
                     if mytype == "integer":
-                        lower, upper = int(lower), int(upper)
+                        lower, upper = int(lower_), int(upper_)
 
                         if keep_value_names:
-                            issues[issue_key] = list(range(lower, upper + 1))
+                            if (upper + 1 - lower) > 1_000_000:
+                                warnings.warn(
+                                    f"Issue {issue_key} has bounds ({lower}, {upper}) which means "
+                                    f"{upper + 1 - lower} values. Consider NOT using keep_value_names"
+                                    f" to reduce memory consumption"
+                                )
+                            issues_dict[issue_key] = list(range(lower, upper + 1))
                         else:
-                            issues[issue_key] = upper - lower + 1
+                            issues_dict[issue_key] = lower, upper
                     else:
+                        lower, upper = float(lower_), float(upper_)
                         if n_discretization is None:
                             all_discrete = False
-                            issues[issue_key] = (float(lower), float(upper))
+                            issues_dict[issue_key] = lower, upper
                         else:
-                            issues[issue_key] = n_discretization
+                            issues_dict[issue_key] = n_discretization
                 else:
-                    # I should add the real-valued issues code here
+                    # I should add the real-valued issues_dict code here
                     raise ValueError(f"Unknown type: {mytype}")
             else:
                 raise ValueError(f"Unknown child for objective: {child.tag}")
 
-        for key, value in zip(ikeys(issues), ivalues(issues)):
-            issues[key] = Issue(
+        for key, value in zip(ikeys(issues_dict), ivalues(issues_dict)):
+            issues_dict[key] = Issue(
                 values=value,
                 name=issue_info[key]["name"]
                 if keep_issue_names
                 else str(issue_info[key]["index"]),
             )
-        issues = list(issues.values())
+        issues = list(issues_dict.values())
 
         if force_single_issue:
             issue_name_ = (
@@ -753,10 +779,10 @@ class Issue(NamedObject):
                 if keep_value_names:
                     if len(issues) > 1:
                         all_values = itertools.product(
-                            *[
+                            *(
                                 [str(_) for _ in issue.alli(n=n_discretization)]
                                 for issue_key, issue in zip(range(len(issues)), issues)
-                            ]
+                            )
                         )
                         all_values = list(
                             map(lambda items: "+".join(items), all_values)
@@ -779,15 +805,15 @@ class Issue(NamedObject):
     @classmethod
     def from_genius(
         cls,
-        file_name: str,
+        file_name: PATH,
         force_single_issue=False,
         force_numeric=False,
         keep_value_names=True,
         keep_issue_names=True,
         safe_parsing=True,
         n_discretization: Optional[int] = None,
-        max_n_outcomes: int = 1e6,
-    ):
+        max_n_outcomes: int = 1_000_000,
+    ) -> Tuple[Optional[List["Issue"]], Optional[List[str]]]:
         """Imports a the domain issues from a GENIUS XML file.
 
         Args:
@@ -803,7 +829,10 @@ class Issue(NamedObject):
 
         Returns:
 
-            A List[Issue] or Dict[Issue]
+            A two optional lists:
+
+            - List[Issue] containing the issues
+            - List[str] containing agent names (that are sometimes stored in the genius domain)
 
 
         Examples:
@@ -818,7 +847,7 @@ class Issue(NamedObject):
             See ``from_xml_str`` for all the parameters
 
         """
-        with open(file_name, "r", encoding="utf-8") as f:
+        with open(file_name, encoding="utf-8") as f:
             xml_str = f.read()
 
             return cls.from_xml_str(
@@ -871,7 +900,7 @@ class Issue(NamedObject):
         for i, issue in enumerate(issues):
             count = 1 if one_each else counts[i]  # type: ignore
 
-            for j in range(count):
+            for _ in range(count):
                 name = str(nxt) if int_names else names[i]  # type: ignore
                 # if count > 1:
                 #    name = name + f' {j}'
@@ -898,6 +927,15 @@ class Issue(NamedObject):
 
         return "uncountable" if self._is_generator else "discrete"
 
+    def is_numeric(self) -> bool:
+        return self._is_integer_valued or self._is_real_valued
+
+    def is_integer(self) -> bool:
+        return self._is_integer_valued
+
+    def is_float(self) -> bool:
+        return self._is_real_valued
+
     def is_continuous(self) -> bool:
         """Test whether the issue is a continuous issue
 
@@ -1116,11 +1154,13 @@ class Issue(NamedObject):
                 fail_if_not_enough=False,
                 with_replacement=False,
             )
-        values = enumerate_outcomes(issues, astype=tuple)
 
         if max_n_outcomes is not None and n > max_n_outcomes:
-            values = random.sample(values, max_n_outcomes)
-
+            values = sample_outcomes(
+                issues=issues, n_outcomes=max_n_outcomes, astype=tuple
+            )
+        else:
+            values = enumerate_outcomes(issues, astype=tuple)
         if issubclass(astype, tuple):
             return values
         issue_names = [_.name for _ in issues]
@@ -1134,6 +1174,7 @@ class Issue(NamedObject):
         issues: Collection["Issue"],
         n_discretization: int = 10,
         astype: Type = dict,
+        max_n_outcomes: int = None,
     ) -> List["Outcome"]:
         """
         Enumerates the outcomes of a list of issues.
@@ -1152,7 +1193,7 @@ class Issue(NamedObject):
             else Issue(values=_.alli(n_discretization), name=_.name)
             for _ in issues
         ]
-        return cls.enumerate(issues, max_n_outcomes=None, astype=astype)
+        return cls.enumerate(issues, max_n_outcomes=max_n_outcomes, astype=astype)
 
     @classmethod
     def sample(
@@ -1219,6 +1260,8 @@ class Issue(NamedObject):
 
 
         """
+        if astype is None:
+            astype = dict
         n_total = num_outcomes(issues)
 
         if (
@@ -1303,6 +1346,50 @@ class Issue(NamedObject):
 
     __repr__ = __str__
 
+    @classmethod
+    def combine(
+        cls,
+        issues: Iterable["Issue"],
+        issue_name="combined",
+        keep_issue_names=True,
+        keep_value_names=True,
+        issue_sep="_",
+        value_sep="-",
+    ) -> Optional["Issue"]:
+        """
+        Combines multiple issues into a single issue
+
+        Args:
+            issues: The issues to be combined
+            issue_name: used only if `keep_issue_names` is False
+            keep_issue_names: If true, the final issue name will be a
+                              concatenation of issue names separated by `issue_sep`
+                              otherwise `issue_name` will be used.
+            keep_value_names: If true, the values for the generated issue
+                              will be a concatenation of values from earlier
+                              issues separated by `value_sep`.
+            issue_sep: Separator for the issue name (used only if `keep_issue_names`)
+            value_sep: Separator for the issue name (used only if `keep_value_names`)
+
+        Remarks:
+
+            - Only works if the issues have finite cardinality
+
+        """
+        n_outcomes = cls.num_outcomes(issues)
+        if n_outcomes == float("inf"):
+            return None
+        if keep_issue_names:
+            issue_name = issue_sep.join([_.name for _ in issues])
+        if keep_value_names:
+            values = [
+                value_sep.join([str(_) for _ in outcome])
+                for outcomes in cls.enumerate(issues, astype=tuple)
+            ]
+        else:
+            values = n_outcomes
+        return Issue(name=issue_name, values=values)
+
     def __hash__(self):
         return hash(str(self))
 
@@ -1329,6 +1416,27 @@ class Issue(NamedObject):
             f"Unknown issue type: {class_name} with dict {d} received from Java"
         )
 
+    def to_dict(self):
+        return dict(
+            values=self._values,
+            name=self.name,
+            id=self.id,
+            value_type=self._value_type,
+            min_value=self.min_value,
+            max_value=self.max_value,
+        )
+
+    @classmethod
+    def from_dict(cls, d):
+        return cls(
+            values=d.get("values", None),
+            name=d.get("name", None),
+            id=d.get("id", None),
+            value_type=d.get("value_type", None),
+            min_value=d.get("min_value", None),
+            max_value=d.get("max_value", None),
+        )
+
     def to_java(self):
         if self._values is None:
             return None
@@ -1373,7 +1481,7 @@ class Issue(NamedObject):
         }
 
 
-class Issues(object):
+class Issues:
     """Encodes a set of Issues.
 
     Args:
@@ -1409,7 +1517,7 @@ class Issues(object):
     @property
     def num_outcomes(self) -> Union[int, float]:
         """Returns the total number of outcomes in a set of issues. Infinity is returned for uncountable or continuous
-        outcomes """
+        outcomes"""
         n = 1
 
         for issue in self.issues:
@@ -1636,7 +1744,7 @@ def enumerate_outcomes(
         )
         astype = dict if keep_issue_names else tuple
     try:
-        outcomes = list(tuple(_) for _ in itertools.product(*[_.all for _ in issues]))
+        outcomes = list(tuple(_) for _ in itertools.product(*(_.all for _ in issues)))
     except:
         return None
 
@@ -1784,8 +1892,28 @@ def sample_outcomes(
                 outcomes += outcomes[:n_rem]
 
             return outcomes
-
-    return list(random.sample(enumerate_outcomes(issues, astype=astype), n_outcomes))
+    n_per_issue = 1 + int(n_outcomes ** (1 / len(issues)))
+    vals = []
+    n_found = 1
+    for issue in issues:
+        if n_per_issue < 2:
+            if random.random() < 0.5:
+                vals.append((issue.min_value,))
+            else:
+                vals.append((issue.max_value,))
+            continue
+        vals.append(issue.alli(n=n_per_issue))
+        n_found *= n_per_issue
+        if n_found > n_outcomes:
+            n_per_issue = 1
+    outcomes = itertools.product(*vals)
+    if issubclass(astype, tuple):
+        return list(outcomes)[:n_outcomes]
+    names = [i.name for i in issues]
+    outcomes = list(dict(zip(names, o)) for o in outcomes)[:n_outcomes]
+    if issubclass(astype, dict):
+        return outcomes
+    return [astype(*o) for o in outcomes]
 
 
 def _is_single(x):
@@ -1847,6 +1975,39 @@ def outcome_is_valid(outcome: "Outcome", issues: Collection[Issue]) -> bool:
     return True
 
 
+def outcome_types_are_ok(outcome: "Outcome", issues: List[Issue]) -> bool:
+    """
+    Checks that the types of all issue values in the outcome are correct
+    """
+    if not issues or not outcome:
+        return True
+    outcome = outcome_as_tuple(outcome)
+    for v, i in zip(outcome, issues):
+        if i.value_type is None:
+            continue
+        if not isinstance(v, i.value_type):
+            return False
+    return True
+
+
+def cast_outcome(outcome: "Outocme", issues: List[Issue]) -> bool:
+    """
+    Casts the types of values in the outcomes to the value-type of each issue (if given)
+    """
+    if not issues or not outcome:
+        return outcome
+    is_dict = isinstance(outcome, dict)
+    if is_dict:
+        keys = outcome.keys()
+        outcome = outcome_as_tuple(outcome)
+    new_outcome = list(outcome)
+    for indx, (v, i) in enumerate(zip(outcome, issues)):
+        if i.value_type is None:
+            continue
+        new_outcome[indx] = i.value_type(v)
+    return dict(zip(keys, new_outcome)) if is_dict else tuple(new_outcome)
+
+
 def outcome_is_complete(outcome: "Outcome", issues: Collection[Issue]) -> bool:
     """Tests that the outcome is valid and complete.
 
@@ -2106,7 +2267,7 @@ def outcome_as_dict(outcome: "Outcome", issue_names: List[str] = None):
 
     if issue_names is not None:
         return dict(zip(issue_names, outcome))
-
+    warnings.warn(f"Outcome {outcome} is converted to a dict without issue names!!")
     return dict(zip((str(_) for _ in range(len(outcome))), outcome))
 
 
