@@ -6,15 +6,14 @@ import random
 from abc import ABC, abstractmethod
 from collections import defaultdict
 from functools import lru_cache
-
 from typing import Dict, List, Optional, Tuple, Union
 
-from .common import SAOResponse, SAOState
-from .negotiators import AspirationNegotiator, PassThroughSAONegotiator, SAONegotiator
 from ..common import AgentMechanismInterface, MechanismState
 from ..negotiators import AspirationMixin, Controller
 from ..outcomes import Outcome, ResponseType, outcome_is_valid
-from ..utilities import UtilityFunction, utility_range
+from ..utilities import utility_range
+from .common import SAOResponse, SAOState
+from .negotiators import AspirationNegotiator, PassThroughSAONegotiator, SAONegotiator
 
 __all__ = [
     "SAOController",
@@ -45,7 +44,7 @@ class SAOController(Controller):
         self,
         default_negotiator_type=PassThroughSAONegotiator,
         default_negotiator_params=None,
-        auto_kill=True,
+        auto_kill=False,
         name=None,
         ufun=None,
     ):
@@ -106,7 +105,7 @@ class SAOController(Controller):
 
     def propose(self, negotiator_id: str, state: MechanismState) -> Optional["Outcome"]:
 
-        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
+        negotiator, _ = self._negotiators.get(negotiator_id, (None, None))
         if negotiator is None:
             raise ValueError(f"Unknown negotiator {negotiator_id}")
         return self.call(negotiator, "propose", state=state)
@@ -121,7 +120,7 @@ class SAOController(Controller):
 
     def on_negotiation_end(self, negotiator_id: str, state: MechanismState) -> None:
         if self._auto_kill:
-            self.kill_negotiator(negotiator_id, True)
+            self.kill_negotiator(negotiator_id)
 
     def on_negotiation_start(self, negotiator_id: str, state: MechanismState) -> None:
         pass
@@ -164,6 +163,11 @@ class SAOSyncController(SAOController):
     """
     A controller that can manage multiple negotiators synchronously.
 
+    Args:
+
+        global_ufun: If true, the controller assumes that the ufun is only
+                     defined globally for the complete set of negotiations
+
     Remarks:
         - The controller waits for an offer from each one of its negotiators before deciding what to do.
         - Loops may happen if multiple controllers of this type negotiate with each other. For example controller A
@@ -173,63 +177,95 @@ class SAOSyncController(SAOController):
 
     """
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args, global_ufun=False, **kwargs):
         super().__init__(*args, **kwargs)
-        self.offers: Dict[str, "Outcome"] = {}
+        self.__offers: Dict[str, "Outcome"] = {}
         """Keeps the last offer received for each negotiation"""
-        self.responses: Dict[str, ResponseType] = {}
+        self.__responses: Dict[str, ResponseType] = {}
         """Keeps the next response type for each negotiation"""
-        self.proposals: Dict[str, "Outcome"] = {}
+        self.__proposals: Dict[str, "Outcome"] = {}
         """Keeps the next proposal for each negotiation"""
-        self.offer_states: Dict[str, "SAOState"] = {}
+        self.__offer_states: Dict[str, "SAOState"] = {}
         """Keeps the last state received for each negotiation"""
-        self.n_waits: Dict[str, int] = defaultdict(int)
+        self.__n_waits: Dict[str, int] = defaultdict(int)
+        """The number of contiguous waits sent for every partner"""
+        self.__global_ufun = global_ufun
 
     def propose(self, negotiator_id: str, state: MechanismState) -> Optional["Outcome"]:
         # if there are no proposals yet, get first proposals
-        if len(self.proposals) == 0:
-            self.proposals = self.first_proposals()
+        if not self.__proposals:
+            self.__proposals = self.first_proposals()
         # get the saved proposal if it exists and return it
-        proposal = self.proposals.get(negotiator_id, None)
-        # if some proposal was there, delete it to force the controller to get a new one
-        if proposal is not None:
-            self.proposals[negotiator_id] = None
-        # if the proposal that was there was None, just offer the best offer
-        proposal = self.first_offer(negotiator_id)
+        if negotiator_id in self.__proposals.keys():
+            # if some proposal was there, delete it to force the controller to get a new one
+            proposal = self.__proposals.pop(negotiator_id)
+        else:
+            # if there was no proposal, get one. Note that `None` is a valid proposal
+            if self.__global_ufun:
+                self.__proposals = self.first_proposals()
+                proposal = self.__proposals.pop(negotiator_id)
+            else:
+                proposal = self.first_offer(negotiator_id)
+
+        # report not waiting on this offer because I obviously just sent
+        self.__n_waits[negotiator_id] = 0
+        self.__responses.pop(negotiator_id, None)
+        self.__offers.pop(negotiator_id, None)
+        self.__offer_states.pop(negotiator_id, None)
         return proposal
 
     def respond(
         self, negotiator_id: str, state: MechanismState, offer: "Outcome"
     ) -> "ResponseType":
         # get the saved response to this negotiator if any
-        response = self.responses.get(negotiator_id, None)
-        if response is not None:
+        response = self.__responses.pop(negotiator_id, ResponseType.WAIT)
+
+        # if there some non-waiting saved respons return and delete it
+        if response != ResponseType.WAIT:
             # remove the response and return it
-            del self.responses[negotiator_id]
-            self.n_waits[negotiator_id] = 0
+            self.__n_waits[negotiator_id] = 0
             return response
 
+        # we get here if there was no saved response (WAIT should never be saved)
+
         # set the saved offer for this negotiator
-        self.offers[negotiator_id] = offer
-        self.offer_states[negotiator_id] = state
+        self.__offers[negotiator_id] = offer
+        self.__offer_states[negotiator_id] = state
         n_negotiators = len(self.active_negotiators)
-        # if we got all the offers or waited long enough, counter all the offers so-far
+        # if we did not get all the offers yet and can still wait, wait
         if (
-            len(self.offers) == n_negotiators
-            or self.n_waits[negotiator_id] >= n_negotiators
+            len(self.__offers) < n_negotiators
+            and self.__n_waits[negotiator_id] < n_negotiators
         ):
-            responses = self.counter_all(offers=self.offers, states=self.offer_states)
-            for nid in responses.keys():
-                # register the responses for next time for all other negotiators
-                if nid != negotiator_id:
-                    self.responses[nid] = responses[nid].response
-                self.proposals[nid] = responses[nid].outcome
-            self.offers = dict()
-            self.offer_states = dict()
-            self.n_waits[negotiator_id] = 0
-            return responses[negotiator_id].response
-        self.n_waits[negotiator_id] += 1
-        return ResponseType.WAIT
+            self.__n_waits[negotiator_id] += 1
+            return ResponseType.WAIT
+
+        # we arrive here if we already have all the offers to counter
+        responses = self.counter_all(offers=self.__offers, states=self.__offer_states)
+        for nid in responses.keys():
+            saved_response = responses.get(nid, None)
+            if saved_response is None:
+                self.__responses[nid] = ResponseType.REJECT_OFFER
+                self.__proposals[nid] = None
+                continue
+            # register the responses for next time for all other negotiators
+            self.__responses[nid] = saved_response.response
+            # register the proposals to be sent to all agents including this one
+            self.__proposals[nid] = saved_response.outcome
+            # register that we are not waiting anymore on any of the offers we received
+            self.__n_waits[nid] = 0
+        self.__offers = dict()
+        self.__offer_states = dict()
+        return self.__responses.pop(negotiator_id)
+
+    def first_proposals(self) -> Dict[str, "Outcome"]:
+        """Gets a set of proposals to use for initializing the negotiation."""
+        return dict(
+            zip(
+                self.negotiators.keys(),
+                (self.first_offer(_) for _ in self.negotiators.keys()),
+            )
+        )
 
     @abstractmethod
     def counter_all(
@@ -248,17 +284,6 @@ class SAOSyncController(SAOController):
 
         """
 
-    def first_proposals(self) -> Dict[str, "Outcome"]:
-        """Gets a set of proposals to use for initializing the negotiation. To avoid offering anything, just return None
-        for all of them. That is the default"""
-        return dict(
-            zip(
-                self.negotiators.keys(),
-                (self.first_offer(_) for _ in self.negotiators.keys()),
-            )
-        )
-
-    @lru_cache(maxsize=100)
     def first_offer(self, negotiator_id: str) -> Optional["Outcome"]:
         """
         Finds the first offer for this given negotiator. By default it will be the best offer
@@ -274,23 +299,34 @@ class SAOSyncController(SAOController):
             defined for the negotiator. If neither exists, the first offer will be None.
         """
         negotiator, _ = self.negotiators.get(negotiator_id, (None, None))
-        if negotiator is None:
+        if negotiator is None or negotiator.ami is None:
             return None
-        if negotiator.ami is None:
-            return None
-        # if the controller has a ufun, use it otherwise use the negotiator ufun
-        if self.ufun is not None:
+        # if the controller has a ufun, use it otherwise use the negotiator's
+        ufun = self.ufun if self.ufun is not None else negotiator.ufun
+        if ufun is not None:
             _, _, _, best = utility_range(
-                self.ufun, issues=negotiator.ami.issues, return_outcomes=True
-            )
-        elif negotiator.ufun is not None:
-            _, _, _, best = utility_range(
-                negotiator.ufun, issues=negotiator.ami.issues, return_outcomes=True
+                ufun,
+                issues=negotiator.ami.issues,
+                return_outcomes=True,
+                ami=negotiator.ami,
             )
         else:
             best = None
         return best
 
+    def on_negotiation_end(self, negotiator_id: str, state: MechanismState) -> None:
+        if negotiator_id in self.__offers.keys():
+            del self.__offers[negotiator_id]
+        if negotiator_id in self.__offer_states.keys():
+            del self.__offer_states[negotiator_id]
+        if negotiator_id in self.__responses.keys():
+            del self.__responses[negotiator_id]
+        if negotiator_id in self.__proposals.keys():
+            del self.__proposals[negotiator_id]
+        if negotiator_id in self.__n_waits.keys():
+            del self.__n_waits[negotiator_id]
+        return super().on_negotiation_end(negotiator_id, state)
+
 
 class SAORandomSyncController(SAOSyncController):
     """
@@ -547,10 +583,8 @@ class SAOSingleAgreementController(SAOSyncController, ABC):
         return self._best_outcomes[negotiator][0]
 
     def after_join(self, negotiator_id, ami, state, *, ufun=None, role="agent"):
-        result = super().join(negotiator_id, ami, state, ufun=ufun, role=role)
-        if result is not None:
-            self._best_outcomes[negotiator_id] = self.best_outcome(negotiator_id)
-        return result
+        super().after_join(negotiator_id, ami, state, ufun=ufun, role=role)
+        self._best_outcomes[negotiator_id] = self.best_outcome(negotiator_id)
 
     def best_outcome(
         self, negotiator: str, state: Optional[SAOState] = None
@@ -586,7 +620,7 @@ class SAOSingleAgreementController(SAOSyncController, ABC):
         if ufun is None:
             return None
         _, _, _, top_outcome = ufun.utility_range(
-            issues=neg.ami.issues, return_outcomes=True
+            issues=neg.ami.issues, return_outcomes=True, ami=neg.ami
         )
         return top_outcome
 
@@ -629,7 +663,7 @@ class SAOSingleAgreementController(SAOSyncController, ABC):
 
     @abstractmethod
     def is_acceptable(self, offer: "Outcome", source: str, state: SAOState) -> bool:
-        """ Should decide if the given offer is acceptable
+        """Should decide if the given offer is acceptable
 
         Args:
             offer: The offer being tested
@@ -814,3 +848,4 @@ class SAOSingleAgreementAspirationController(
         )
         if outcome is None:
             return self._best_outcomes.get(negotiator, None)
+        return outcome
