@@ -5,17 +5,10 @@ import math
 import random
 import warnings
 from abc import abstractmethod
-from typing import Callable, List, Optional, Tuple, Type, Union
+from typing import List, Optional, Type, Union
 
 import numpy as np
 
-from .common import SAOResponse
-from .components import (
-    LimitedOutcomesAcceptorMixin,
-    LimitedOutcomesMixin,
-    RandomProposalMixin,
-    RandomResponseMixin,
-)
 from ..common import (
     AgentMechanismInterface,
     MechanismState,
@@ -38,10 +31,16 @@ from ..utilities import (
     LinearUtilityFunction,
     MappingUtilityFunction,
     UtilityFunction,
-    UtilityValue,
     outcome_with_utility,
     utility_range,
 )
+from .common import SAOResponse
+from .components import (
+    LimitedOutcomesAcceptorMixin,
+    LimitedOutcomesMixin,
+    RandomProposalMixin,
+    RandomResponseMixin,
+)
 
 __all__ = [
     "SAONegotiator",
@@ -85,11 +84,12 @@ class SAONegotiator(Negotiator):
         assume_normalized=True,
         ufun: Optional[UtilityFunction] = None,
         name: Optional[str] = None,
-        rational_proposal=True,
         parent: Controller = None,
         owner: "Agent" = None,
+        id: Optional[str] = None,
+        rational_proposal=True,
     ):
-        super().__init__(name=name, ufun=ufun, parent=parent, owner=owner)
+        super().__init__(name=name, ufun=ufun, parent=parent, owner=owner, id=id)
         self.assume_normalized = assume_normalized
         self.__end_negotiation = False
         self.my_last_proposal: Optional["Outcome"] = None
@@ -510,7 +510,8 @@ class AspirationNegotiator(SAONegotiator, AspirationMixin):
         if assume_normalized:
             self.ufun_max, self.ufun_min = 1.0, 0.0
         super().__init__(
-            assume_normalized=assume_normalized, **kwargs,
+            assume_normalized=assume_normalized,
+            **kwargs,
         )
         self.aspiration_init(
             max_aspiration=max_aspiration, aspiration_type=aspiration_type
@@ -538,6 +539,9 @@ class AspirationNegotiator(SAONegotiator, AspirationMixin):
 
     def on_ufun_changed(self):
         super().on_ufun_changed()
+        if self.ufun is None or self._ami is None:
+            self.ufun_max = self.ufun_min = None
+            return
         presort = self.presort
         if (
             not presort
@@ -603,7 +607,11 @@ class AspirationNegotiator(SAONegotiator, AspirationMixin):
     def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
         if self.ufun_max is None or self.ufun_min is None:
             self.on_ufun_changed()
-        if self._utility_function is None:
+        if (
+            self._utility_function is None
+            or self.ufun_max is None
+            or self.ufun_min is None
+        ):
             return ResponseType.REJECT_OFFER
         u = self._utility_function(offer)
         if u is None or u < self.reserved_value:
@@ -621,6 +629,12 @@ class AspirationNegotiator(SAONegotiator, AspirationMixin):
     def propose(self, state: MechanismState) -> Optional["Outcome"]:
         if self.ufun_max is None or self.ufun_min is None:
             self.on_ufun_changed()
+        if (
+            self._utility_function is None
+            or self.ufun_max is None
+            or self.ufun_min is None
+        ):
+            return None
         if self.ufun_max < self.reserved_value:
             return None
         asp = (
@@ -632,7 +646,7 @@ class AspirationNegotiator(SAONegotiator, AspirationMixin):
         if self.presorted:
             if len(self.ordered_outcomes) < 1:
                 return None
-            for i, (u, o) in enumerate(self.ordered_outcomes):
+            for i, (u, _) in enumerate(self.ordered_outcomes):
                 if u is None:
                     continue
                 if u < asp:
@@ -660,7 +674,9 @@ class AspirationNegotiator(SAONegotiator, AspirationMixin):
                 rng = self.ufun_max - self.ufun_min
                 mx = min(asp + tol * rng, self.__last_offer_util)
                 outcome = outcome_with_utility(
-                    ufun=self._utility_function, rng=(asp, mx), issues=self._ami.issues,
+                    ufun=self._utility_function,
+                    rng=(asp, mx),
+                    issues=self._ami.issues,
                 )
                 if outcome is not None:
                     break
@@ -753,7 +769,9 @@ class ToughNegotiator(SAONegotiator):
         if self._utility_function is None:
             return
         _, _, _, self.best_outcome = utility_range(
-            self.utility_function, issues=self.ami.issues, return_outcomes=True,
+            self.utility_function,
+            issues=self.ami.issues,
+            return_outcomes=True,
         )
 
     def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
@@ -804,7 +822,7 @@ class OnlyBestNegotiator(SAONegotiator):
         self.ordered_outcomes = []
         self.acceptable_outcomes = []
         self.wheel = np.array([])
-        self.offered = set([])
+        self.offered = set()
         super().__init__(name=name, parent=parent, ufun=ufun, **kwargs)
         if not dynamic_ufun:
             warnings.warn(
@@ -946,7 +964,7 @@ class NaiveTitForTatNegotiator(SAONegotiator):
         super().on_ufun_changed()
         outcomes = self._ami.discrete_outcomes()
         self.ordered_outcomes = sorted(
-            [(self._utility_function(outcome), outcome) for outcome in outcomes],
+            ((self._utility_function(outcome), outcome) for outcome in outcomes),
             key=lambda x: x[0],
             reverse=True,
         )
@@ -961,7 +979,7 @@ class NaiveTitForTatNegotiator(SAONegotiator):
             self.received_utilities[0] = self.received_utilities[1]
             self.received_utilities[-1] = offered_utility
         indx = self._propose(state=state)
-        my_utility, my_offer = self.ordered_outcomes[indx]
+        my_utility, _ = self.ordered_outcomes[indx]
         if offered_utility >= my_utility:
             return ResponseType.ACCEPT_OFFER
         return ResponseType.REJECT_OFFER
@@ -1040,14 +1058,24 @@ class PassThroughSAONegotiator(SAONegotiator):
         if self._Negotiator__parent:
             return self._Negotiator__parent.on_negotiation_end(self.id, state)
 
-    def join(self, ami, state, *, ufun=None, role="agent",) -> bool:
+    def join(
+        self,
+        ami,
+        state,
+        *,
+        ufun=None,
+        role="agent",
+    ) -> bool:
         """
         Joins a negotiation.
 
         Remarks:
-            This method first gets permission from the parent controller by calling `before_join` on it and confirming
-            the result is `True`, it then joins the negotiation and calls `after_join` of the controller to inform it
-            that joining is completed if joining was successful.
+
+            This method first gets permission from the parent controller by
+            calling `before_join` on it and confirming the result is `True`,
+            it then joins the negotiation and calls `after_join` of the
+            controller to inform it that joining is completed if joining was
+            successful.
         """
         permission = (
             self._Negotiator__parent is None
@@ -1210,7 +1238,7 @@ class JavaSAONegotiator(SAONegotiator, JavaCallerMixin):
 
     @classmethod
     def from_dict(
-        cls, java_object, *args, parent: Controller = None
+        cls, java_object, *args, parent: Controller = None, **kwargs
     ) -> "JavaSAONegotiator":
         """Creates a Java negotiator from an object returned from the JVM implementing PySAONegotiator"""
         ufun = java_object.getUtilityFunction()
