@@ -16,9 +16,11 @@ import yaml
 from tabulate import tabulate
 
 import negmas
-from negmas.helpers import humanize_time, load, unique_name
+from negmas.genius.common import DEFAULT_JAVA_PORT
+from negmas.helpers import humanize_time, load, truncated_mean, unique_name
 from negmas.java import init_jnegmas_bridge, jnegmas_bridge_is_running
 from negmas.tournaments import (
+    combine_tournament_results,
     combine_tournament_stats,
     combine_tournaments,
     create_tournament,
@@ -109,200 +111,6 @@ def tournament(ctx, ignore_warnings):
     ctx.obj = {}
 
 
-# @cli.group(invoke_without_command=True)
-# @click.option(
-#     "--name",
-#     "-n",
-#     default="random",
-#     help='The name of the tournament. The special value "random" will result in a random name',
-# )
-# @click.option("--steps", "-s", default=100, help="Number of steps.")
-# @click.option(
-#     "--ttype",
-#     "--tournament-type",
-#     "--tournament",
-#     default="anac2019collusion",
-#     type=click.Choice(["anac2019collusion", "anac2019std", "anac2019sabotage"]),
-#     help="The config to use. Default is ANAC 2019. Options supported are anac2019std, anac2019collusion, "
-#     "anac2019sabotage",
-# )
-# @click.option(
-#     "--timeout",
-#     "-t",
-#     default=0,
-#     type=int,
-#     help="Timeout the whole tournament after the given number of seconds (0 for infinite)",
-# )
-# @click.option(
-#     "--configs",
-#     default=5,
-#     type=int,
-#     help="Number of unique configurations to generate.",
-# )
-# @click.option("--runs", default=2, help="Number of runs for each configuration")
-# @click.option(
-#     "--max-runs",
-#     default=-1,
-#     type=int,
-#     help="Maximum total number of runs. Zero or negative numbers mean no limit",
-# )
-# @click.option(
-#     "--agents",
-#     default=5,
-#     type=int,
-#     help="Number of agents per competitor (not used for anac2019std in which this is preset to 1).",
-# )
-# @click.option(
-#     "--factories",
-#     default=5,
-#     type=int,
-#     help="Minimum numbers of factories to have per level.",
-# )
-# @click.option(
-#     "--competitors",
-#     default="DoNothingFactoryManager;GreedyFactoryManager",
-#     help="A semicolon (;) separated list of agent types to use for the competition.",
-# )
-# @click.option(
-#     "--jcompetitors",
-#     "--java-competitors",
-#     default="",
-#     help="A semicolon (;) separated list of agent types to use for the competition.",
-# )
-# @click.option(
-#     "--non-competitors",
-#     default="",
-#     help="A semicolon (;) separated list of agent types to exist in the worlds as non-competitors "
-#     "(their scores will not be calculated).",
-# )
-# @click.option(
-#     "--parallel/--serial",
-#     default=True,
-#     help="Run a parallel/serial tournament on a single machine",
-# )
-# @click.option(
-#     "--distributed/--single-machine",
-#     default=False,
-#     help="Run a distributed tournament using dask",
-# )
-# @click.option(
-#     "--log",
-#     "-l",
-#     type=click.Path(dir_okay=True, file_okay=False),
-#     default="~/negmas/logs/tournaments",
-#     help="Default location to save logs (A folder will be created under it)",
-# )
-# @click.option(
-#     "--verbosity",
-#     default=1,
-#     type=int,
-#     help="verbosity level (from 0 == silent to 1 == world progress)",
-# )
-# @click.option("--configs-only/--run", default=False, help="configs_only")
-# @click.option(
-#     "--reveal-names/--hidden-names",
-#     default=True,
-#     help="Reveal agent names (should be used only for " "debugging)",
-# )
-# @click.option(
-#     "--ip",
-#     default="127.0.0.1",
-#     help="The IP address for a dask scheduler to run the distributed tournament."
-#     " Effective only if --distributed",
-# )
-# @click.option(
-#     "--port",
-#     default=8786,
-#     type=int,
-#     help="The IP port number a dask scheduler to run the distributed tournament."
-#     " Effective only if --distributed",
-# )
-# @click.option(
-#     "--log-ufuns/--no-ufun-logs",
-#     default=False,
-#     help="Log ufuns into their own CSV file. Only effective if --debug is given",
-# )
-# @click.option(
-#     "--log-negs/--no-neg-logs",
-#     default=False,
-#     help="Log all negotiations. Only effective if --debug is given",
-# )
-# @click.option(
-#     "--compact/--debug",
-#     default=True,
-#     help="If True, effort is exerted to reduce the memory footprint which"
-#     "includes reducing logs dramatically.",
-# )
-# @click.option(
-#     "--raise-exceptions/--ignore-exceptions",
-#     default=True,
-#     help="Whether to ignore agent exceptions",
-# )
-# @click.option(
-#     "--path",
-#     default="",
-#     help="A path to be added to PYTHONPATH in which all competitors are stored. You can path a : separated list of "
-#     "paths on linux/mac and a ; separated list in windows",
-# )
-# @click_config_file.configuration_option()
-# def tournament(
-#     name,
-#     steps,
-#     parallel,
-#     distributed,
-#     ttype,
-#     timeout,
-#     log,
-#     verbosity,
-#     configs_only,
-#     reveal_names,
-#     ip,
-#     port,
-#     runs,
-#     configs,
-#     max_runs,
-#     competitors,
-#     jcompetitors,
-#     non_competitors,
-#     compact,
-#     factories,
-#     agents,
-#     log_ufuns,
-#     log_negs,
-#     path,
-#     raise_exceptions,
-# ):
-#     if name == "random":
-#         name = unique_name("", rand_digits=0)
-#     create(
-#         name,
-#         steps,
-#         parallel,
-#         distributed,
-#         ttype,
-#         timeout,
-#         log,
-#         verbosity,
-#         reveal_names,
-#         ip,
-#         port,
-#         runs,
-#         configs,
-#         max_runs,
-#         competitors,
-#         jcompetitors,
-#         non_competitors,
-#         compact,
-#         factories,
-#         agents,
-#         log_ufuns,
-#         log_negs,
-#         raise_exceptions,
-#     )
-#     if not configs_only:
-#         run(name, verbosity, distributed, compact, path, log)
-
-
 @tournament.command(help="Creates a tournament")
 @click.option(
     "--name",
@@ -343,7 +151,10 @@ def tournament(ctx, ignore_warnings):
     help="Maximum number of steps (only used if --steps was not passed",
 )
 @click.option(
-    "--agents", default=3, type=int, help="Number of agents per competitor",
+    "--agents",
+    default=3,
+    type=int,
+    help="Number of agents per competitor",
 )
 @click.option(
     "--competitors",
@@ -737,14 +548,36 @@ def create(
 )
 @click.option(
     "--metric",
-    default="median",
+    default="truncated_mean",
     type=str,
-    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum",
+    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum, truncated_mean",
+)
+@click.option(
+    "--significance/--no-significance",
+    default=False,
+    help="Whether to show significance table",
+)
+@click.option(
+    "--eval/--no-eval",
+    default=False,
+    help="Whether evaluate and show results after the tournament is run",
 )
 @click_config_file.configuration_option()
 @click.pass_context
 def run(
-    ctx, name, verbosity, parallel, distributed, ip, port, compact, path, log, metric
+    ctx,
+    name,
+    verbosity,
+    parallel,
+    distributed,
+    ip,
+    port,
+    compact,
+    path,
+    log,
+    metric,
+    significance,
+    eval,
 ):
     if len(name) == 0:
         name = ctx.obj.get("tournament_name", "")
@@ -776,13 +609,49 @@ def run(
         print_exceptions=verbosity > 1,
     )
     end_time = humanize_time(perf_counter() - start)
-    results = evaluate_tournament(
-        tournament_path=tpath, verbose=verbosity > 0, metric=metric
-    )
-    display_results(results, metric)
+    if eval:
+        results = evaluate_tournament(
+            tournament_path=tpath, verbose=verbosity > 0, metric=metric, compile=True
+        )
+        display_results(results, metric, significance)
     print(f"Finished in {end_time}")
 
 
+@tournament.command(help="Evaluates a tournament and returns the results")
+@click.argument(
+    "path",
+    type=click.Path(dir_okay=True, file_okay=False),
+)
+@click.option(
+    "--metric",
+    default="truncated_mean",
+    type=str,
+    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum, truncated_mean",
+)
+@click.option(
+    "--significance/--no-significance",
+    default=False,
+    help="Whether to show significance table",
+)
+@click.option(
+    "--compile/--show",
+    default=True,
+    help="Whether to recompile results from individual world runs or just show the already-compiled results",
+)
+@click.option(
+    "--verbose/--silent",
+    default=True,
+    help="Whether to be verbose",
+)
+@click_config_file.configuration_option()
+@click.pass_context
+def eval(ctx, path, metric, significance, compile, verbose):
+    results = evaluate_tournament(
+        tournament_path=path, metric=metric, compile=compile, verbose=verbose
+    )
+    display_results(results, metric, significance)
+
+
 @tournament.command(
     help="Finds winners of a tournament or a set of tournaments sharing a root"
 )
@@ -806,13 +675,28 @@ def run(
 )
 @click.option(
     "--metric",
-    default="median",
+    default="truncated_mean",
     type=str,
-    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum",
+    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum, truncated_mean",
+)
+@click.option(
+    "--verbose/--silent",
+    default=True,
+    help="Whether to be verbose",
+)
+@click.option(
+    "--significance/--no-significance",
+    default=False,
+    help="Whether to show significance table",
+)
+@click.option(
+    "--compile/--show",
+    default=True,
+    help="Whether to recompile results from individual world runs or just show the already-compiled results",
 )
 @click_config_file.configuration_option()
 @click.pass_context
-def winners(ctx, name, log, recursive, metric):
+def winners(ctx, name, log, recursive, metric, significance, compile, verbose):
     if len(name) == 0:
         if not recursive:
             name = ctx.obj.get("tournament_name", "")
@@ -834,9 +718,13 @@ def winners(ctx, name, log, recursive, metric):
     else:
         tpath = str(pathlib.Path(log))
     results = evaluate_tournament(
-        tournament_path=tpath, verbose=True, recursive=recursive, metric=metric
+        tournament_path=tpath,
+        verbose=verbose,
+        recursive=recursive,
+        metric=metric,
+        compile=compile,
     )
-    display_results(results, metric)
+    display_results(results, metric, significance)
 
 
 def _path(path) -> Path:
@@ -853,36 +741,92 @@ def _path(path) -> Path:
     return Path(path).absolute()
 
 
-def display_results(results, metric):
-    viewmetric = ["50%" if metric == "median" else metric]
-    print(
-        tabulate(
-            results.score_stats.sort_values(by=viewmetric, ascending=False),
-            headers="keys",
-            tablefmt="psql",
+def display_results(results, metric, significance):
+    if metric == "truncated_mean":
+        print(
+            tabulate(
+                results.total_scores.sort_values(by="score", ascending=False),
+                headers="keys",
+                tablefmt="psql",
+            )
+        )
+        print(
+            tabulate(
+                results.score_stats.sort_values(by="median", ascending=False),
+                headers="keys",
+                tablefmt="psql",
+            )
         )
-    )
-
-    if metric in ("mean", "sum"):
-        print(tabulate(results.ttest, headers="keys", tablefmt="psql"))
     else:
-        print(tabulate(results.kstest, headers="keys", tablefmt="psql"))
+        viewmetric = ["50%" if metric == "median" else metric]
+        print(
+            tabulate(
+                results.score_stats.sort_values(by=viewmetric, ascending=False),
+                headers="keys",
+                tablefmt="psql",
+            )
+        )
 
-    agg_stats = results.agg_stats.loc[
-        :,
-        [
-            "n_negotiations_sum",
-            "n_contracts_concluded_sum",
-            "n_contracts_signed_sum",
-            "n_contracts_executed_sum",
-            "activity_level_sum",
-        ],
-    ]
-    agg_stats.columns = ["negotiated", "concluded", "signed", "executed", "business"]
-    print(tabulate(agg_stats.describe(), headers="keys", tablefmt="psql"))
+    if significance:
+        if metric in ("mean", "sum", "tuncated_mean"):
+            print(tabulate(results.ttest, headers="keys", tablefmt="psql"))
+        else:
+            print(tabulate(results.kstest, headers="keys", tablefmt="psql"))
+
+    try:
+        agg_stats = results.agg_stats.loc[
+            :,
+            [
+                "n_negotiations_sum",
+                "n_contracts_concluded_sum",
+                "n_contracts_signed_sum",
+                "n_contracts_executed_sum",
+                "activity_level_sum",
+            ],
+        ]
+        agg_stats.columns = [
+            "negotiated",
+            "concluded",
+            "signed",
+            "executed",
+            "business",
+        ]
+        print(tabulate(agg_stats.describe(), headers="keys", tablefmt="psql"))
+    except:
+        pass
 
 
-@tournament.command(help="Finds winners of an arbitrary set of tournaments")
+@tournament.command(help="Combine multiple tournaments at the given base path(s)")
+@click.argument(
+    "path",
+    type=click.Path(dir_okay=True, file_okay=False),
+    nargs=-1,
+)
+@click.option(
+    "--verbose/--silent",
+    default=True,
+    help="Whether to be verbose",
+)
+@click.option(
+    "--dest",
+    "-d",
+    type=click.Path(dir_okay=True, file_okay=False),
+    help="The location to save the results",
+    default=None,
+)
+@click_config_file.configuration_option()
+def combine(path, dest, verbose):
+    if dest is None:
+        print("Must specify the destination using --dest/-d option")
+        return
+    tpath = [_path(_) for _ in path]
+
+    if len(tpath) < 1:
+        print("No paths are given to combine")
+    combine_tournaments(sources=tpath, dest=dest, verbose=verbose)
+
+
+@tournament.command(help="Combine results from multiple tournaments")
 @click.argument("path", type=click.Path(dir_okay=True, file_okay=False), nargs=-1)
 @click.option(
     "--dest",
@@ -893,20 +837,37 @@ def display_results(results, metric):
 )
 @click.option(
     "--metric",
-    default="mean",
+    default="truncated_mean",
     type=str,
-    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum",
+    help="The statistical metric used for choosing the winners. Possibilities are mean, median, std, var, sum, truncated_mean",
+)
+@click.option(
+    "--significance/--no-significance",
+    default=False,
+    help="Whether to show significance table",
+)
+@click.option(
+    "--verbose/--silent",
+    default=True,
+    help="Whether to be verbose",
+)
+@click.option(
+    "--compile/--show",
+    default=True,
+    help="Whether to recompile results from individual world runs or just show the already-compiled results",
 )
 @click_config_file.configuration_option()
-def combine(path, dest, metric):
+def combine_results(path, dest, metric, significance, compile, verbose):
     tpath = [_path(_) for _ in path]
 
     if len(tpath) < 1:
         print("No paths are given to combine")
-    scores = combine_tournaments(sources=tpath, dest=None, verbose=True)
-    stats = combine_tournament_stats(sources=tpath, dest=None, verbose=True)
-    results = evaluate_tournament(dest, scores, stats, verbose=True, metric=metric)
-    display_results(results, metric)
+    scores = combine_tournament_results(sources=tpath, dest=None, verbose=verbose)
+    stats = combine_tournament_stats(sources=tpath, dest=None, verbose=verbose)
+    results = evaluate_tournament(
+        dest, scores, stats, verbose=verbose, metric=metric, compile=compile
+    )
+    display_results(results, metric, significance)
 
 
 @cli.command(help="Start the bridge to genius (to use GeniusNegotiator)")
@@ -923,18 +884,32 @@ def combine(path, dest, metric):
 @click.option(
     "--port",
     "-r",
-    default=0,
+    default=DEFAULT_JAVA_PORT,
     help="Port to run the NegLoader on. Pass 0 for the default value",
 )
 @click.option(
-    "--force/--no-force",
+    "--debug/--silent",
     default=False,
-    help="Force trial even if an earlier instance exists",
+    help="Run the bridge in debug mode if --debug else silently",
 )
-def genius(path, port, force):
+@click.option(
+    "--timeout",
+    default=0,
+    type=float,
+    help="The timeout to pass. Zero or negative numbers to disable and use the bridge's global timeout.",
+)
+def genius(path, port, debug, timeout):
+    if port and negmas.genius_bridge_is_running(port):
+        print(f"Genius Bridge is already running on port {port} ... exiting")
+        sys.exit()
     negmas.init_genius_bridge(
-        path=path if path != "auto" else None, port=port, force=force
+        path=path if path != "auto" else None,
+        port=port,
+        debug=debug,
+        timeout=timeout,
     )
+    while True:
+        pass
 
 
 @cli.command(help="Start the bridge to JNegMAS (to use Java agents in worlds)")
@@ -983,7 +958,7 @@ def download_and_set(key, url, file_name):
     config_file = config_path / "config.json"
 
     if config_file.exists():
-        with open(config_file, "r") as f:
+        with open(config_file) as f:
             config = json.load(f)
     config[key] = str(jar_path)
     with open(config_file, "w") as f:
