@@ -7,7 +7,6 @@ import pytest
 
 from negmas import load_genius_domain_from_folder
 from negmas.elicitation import (
-    BalancedElicitor,
     DummyElicitor,
     EStrategy,
     FullKnowledgeElicitor,
@@ -16,12 +15,21 @@ from negmas.elicitation import (
     User,
     next_query,
     possible_queries,
-    OptimalIncrementalElicitor,
 )
-from negmas.helpers import Distribution as U, instantiate
+from negmas.helpers import instantiate
 from negmas.sao import AspirationNegotiator, LimitedOutcomesNegotiator, SAOMechanism
 from negmas.utilities import IPUtilityFunction, MappingUtilityFunction, pareto_frontier
 
+try:
+    from blist import sortedlist
+
+    BLIST_AVAILABLE = True
+except ImportError:
+    BLIST_AVAILABLE = False
+    print(
+        "blist is not avialable. This is a known issue with python 3.9. Use python 3.8 if you are testing VOI"
+    )
+
 n_outcomes = 5
 cost = 0.02
 utility = 0.17682
@@ -33,12 +41,15 @@ ufun = MappingUtilityFunction(
 
 import negmas.elicitation as elicitation
 
-
 all_countable_queries_elicitor_types = [
     _
     for _ in elicitation.__all__
     if _.endswith("Elicitor") and not _.startswith("Base") and not "VOIOptimal" in _
 ]
+if not BLIST_AVAILABLE:
+    all_countable_queries_elicitor_types = [
+        _ for _ in all_countable_queries_elicitor_types if "voi" not in _.lower()
+    ]
 
 
 @pytest.fixture
@@ -76,7 +87,7 @@ def master(true_utilities, strategy_name="titration-0.05"):
     return user, strategy
 
 
-class TestCountableOutcomesUser(object):
+class TestCountableOutcomesUser:
     def test_countable_outcmoes_user_initializable(self):
         user = User(ufun=ufun, cost=cost)
         assert user.total_cost == 0.0, "total cost is not initialized to zero"
@@ -406,7 +417,7 @@ def data_folder():
     return pkg_resources.resource_filename("negmas", resource_name="tests/data")
 
 
-class TestCountableOutcomesElicitor(object):
+class TestCountableOutcomesElicitor:
     def test_dummy(self, master, true_utilities):
         user, strategy = master
         neg = SAOMechanism(outcomes=[(_,) for _ in range(n_outcomes)], n_steps=10)
@@ -483,7 +494,7 @@ class TestCountableOutcomesElicitor(object):
                 == true_utilities[neg.agreement[0]] - elicitor.elicitation_cost
             )
         if hasattr(elicitor, "each_outcome_once") and elicitor.each_outcome_once:
-            assert len(set([_[0] for _ in elicitor.offers])) == len(elicitor.offers)
+            assert len({_[0] for _ in elicitor.offers}) == len(elicitor.offers)
         # print(
         #     f"Got {elicitor.ufun(neg.agreement)} with elicitation cost {elicitor.elicitation_cost} "
         #     f"for {elicitor} using {len(queries)} elicited_queries"
@@ -570,6 +581,7 @@ class TestCountableOutcomesElicitor(object):
             keep_issue_names=True,
             keep_value_names=True,
             agent_factories=lambda: AspirationNegotiator(),
+            normalize_utilities=True,
         )
         # [domain.add(LimitedOutcomesNegotiator(outcomes=n_outcomes)
         #            , ufun=_['ufun']) for _ in agents_info]
@@ -588,6 +600,7 @@ class TestCountableOutcomesElicitor(object):
             keep_issue_names=False,
             keep_value_names=False,
             agent_factories=lambda: AspirationNegotiator(),
+            normalize_utilities=True,
         )
         # [domain.add(LimitedOutcomesNegotiator(outcomes=n_outcomes)
         #            , ufun=_['ufun']) for _ in agents_info]
@@ -624,6 +637,7 @@ class TestCountableOutcomesElicitor(object):
             force_single_issue=True,
             keep_issue_names=False,
             keep_value_names=False,
+            normalize_utilities=True,
         )
         domain.add(LimitedOutcomesNegotiator(), ufun=agents_info[0]["ufun"])
         # domain.n_steps = 10
@@ -640,7 +654,8 @@ class TestCountableOutcomesElicitor(object):
         outcomes = [(0,), (1,)]
         alphas = [0.0, 1.0]
         opponent = LimitedOutcomesNegotiator(
-            acceptable_outcomes=outcomes, acceptance_probabilities=alphas,
+            acceptable_outcomes=outcomes,
+            acceptance_probabilities=alphas,
         )
 
 
@@ -723,6 +738,10 @@ def test_alternating_offers_eliciting_mechanism_voi():
         )
 
 
+@pytest.mark.skipif(
+    condition=not BLIST_AVAILABLE,
+    reason="Blist is not available. Ignoring this test",
+)
 def test_alternating_offers_eliciting_mechanism_voi_optimal():
     config = SAOElicitingMechanism.generate_config(
         cost=0.001,
@@ -788,8 +807,8 @@ def test_a_small_elicitation_session():
 
 
 def test_a_typical_elicitation_session():
-    import random
     import os
+    import random
 
     n_outcomes = 10
     accepted_outcomes = [int(random.random() <= 0.5) for _ in range(n_outcomes)]
