@@ -3,7 +3,7 @@ from typing import Any, Callable, Collection, Dict, List, Optional, Set
 
 import hypothesis.strategies as st
 import pytest
-from hypothesis import given, settings
+from hypothesis import HealthCheck, given, settings
 
 from negmas import (
     AgentMechanismInterface,
@@ -287,7 +287,7 @@ def test_world_has_times(capsys):
 
 
 def test_world_records_exceptions(capsys):
-    import time
+    pass
 
     world = DummyWorld(n_steps=10, ignore_agent_exceptions=True)
     world.join(ExceptionAgent("A1"))
@@ -368,7 +368,7 @@ def test_world_picklable(tmp_path):
     w.run()
 
 
-@settings(deadline=None)
+@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
 @given(
     single_checkpoint=st.booleans(),
     checkpoint_every=st.integers(0, 6),
